@@ -22,8 +22,8 @@
 #include "branch_info.h"
 #include "code_coverage_private.h"
 
+#include "base/base.h"
 #include "base/filter.h"
-#include "base/stack.h"
 #include "lib/compat.h"
 #include "lib/set.h"
 #include "lib/var.h"
@@ -42,12 +42,12 @@ static void xdebug_coverage_line_dtor(void *data)
 	xdfree(line);
 }
 
-xdebug_coverage_file *xdebug_coverage_file_ctor(char *filename)
+xdebug_coverage_file *xdebug_coverage_file_ctor(zend_string *filename)
 {
 	xdebug_coverage_file *file;
 
 	file = xdmalloc(sizeof(xdebug_coverage_file));
-	file->name = xdstrdup(filename);
+	file->name = zend_string_copy(filename);
 	file->lines = xdebug_hash_alloc(128, xdebug_coverage_line_dtor);
 	file->functions = xdebug_hash_alloc(128, xdebug_coverage_function_dtor);
 	file->has_branch_info = 0;
@@ -61,7 +61,7 @@ static void xdebug_coverage_file_dtor(void *data)
 
 	xdebug_hash_destroy(file->lines);
 	xdebug_hash_destroy(file->functions);
-	xdfree(file->name);
+	zend_string_release(file->name);
 	xdfree(file);
 }
 
@@ -87,37 +87,81 @@ void xdebug_coverage_function_dtor(void *data)
 	xdfree(function);
 }
 
-static char* xdebug_func_format(xdebug_func *func)
+static void xdebug_func_format(char *buffer, size_t buffer_size, xdebug_func *func)
 {
-	switch (func->type) {
-		case XFUNC_NORMAL:
-			return xdstrdup(func->function);
-		case XFUNC_MEMBER:
-			return xdebug_sprintf("%s->%s", func->class, func->function);
-		default:
-			return xdstrdup("???");
+	if (func->type == XFUNC_NORMAL) {
+		int len = strlen(func->function);
+
+		if (len + 1 > buffer_size) {
+			goto error;
+		}
+		memcpy(buffer, func->function, len);
+		buffer[len] = '\0';
+		return;
+	}
+
+	if (func->type == XFUNC_MEMBER) {
+		int func_len = strlen(func->function);
+		int len = ZSTR_LEN(func->class_name) + 2 + func_len;
+
+		if (len + 1 > buffer_size) {
+			goto error;
+		}
+		memcpy(buffer, ZSTR_VAL(func->class_name), ZSTR_LEN(func->class_name));
+		memcpy(buffer + ZSTR_LEN(func->class_name), "->", 2);
+		memcpy(buffer + ZSTR_LEN(func->class_name) + 2, func->function, func_len);
+		buffer[len] = '\0';
+		return;
+	}
+
+error:
+	memcpy(buffer, "?", 1);
+	buffer[1] = '\0';
+}
+
+static void xdebug_build_fname_from_oparray(xdebug_func *tmp, zend_op_array *opa)
+{
+	int closure = 0;
+
+	memset(tmp, 0, sizeof(xdebug_func));
+
+	if (opa->function_name) {
+		if (opa->fn_flags & ZEND_ACC_CLOSURE) {
+			tmp->function = xdebug_wrap_closure_location_around_function_name(opa, STR_NAME_VAL(opa->function_name));
+			closure = 1;
+		} else {
+			tmp->function = xdstrdup(STR_NAME_VAL(opa->function_name));
+		}
+	} else {
+		tmp->function = xdstrdup("{main}");
+		tmp->type = XFUNC_MAIN;
+	}
+
+	if (opa->scope && !closure) {
+		tmp->type = XFUNC_MEMBER;
+		tmp->class_name = zend_string_copy(opa->scope->name);
+	} else {
+		tmp->type = XFUNC_NORMAL;
 	}
 }
 
-static void xdebug_print_opcode_info(char type, zend_execute_data *execute_data, const zend_op *cur_opcode)
+static void xdebug_print_opcode_info(zend_execute_data *execute_data, const zend_op *cur_opcode)
 {
 	zend_op_array *op_array = &execute_data->func->op_array;
-	char *file = (char*) STR_NAME_VAL(op_array->filename);
 	xdebug_func func_info;
-	char *function_name;
+	char function_name[1024];
 	long opnr = execute_data->opline - execute_data->func->op_array.opcodes;
 
 	xdebug_build_fname_from_oparray(&func_info, op_array);
-	function_name = xdebug_func_format(&func_info);
-	if (func_info.class) {
-		xdfree(func_info.class);
+	xdebug_func_format(function_name, sizeof(function_name), &func_info);
+	if (func_info.class_name) {
+		zend_string_release(func_info.class_name);
 	}
 	if (func_info.function) {
 		xdfree(func_info.function);
 	}
 
-	xdebug_branch_info_mark_reached(file, function_name, op_array, opnr);
-	xdfree(function_name);
+	xdebug_branch_info_mark_reached(op_array->filename, function_name, op_array, opnr);
 }
 
 static int xdebug_check_branch_entry_handler(XDEBUG_OPCODE_HANDLER_ARGS)
@@ -126,28 +170,31 @@ static int xdebug_check_branch_entry_handler(XDEBUG_OPCODE_HANDLER_ARGS)
 	const zend_op *cur_opcode = execute_data->opline;
 
 	if (!op_array->reserved[XG_COV(code_coverage_filter_offset)] && XG_COV(code_coverage_active)) {
-		xdebug_print_opcode_info('G', execute_data, cur_opcode);
+		xdebug_print_opcode_info(execute_data, cur_opcode);
 	}
 
 	return xdebug_call_original_opcode_handler_if_set(cur_opcode->opcode, XDEBUG_OPCODE_HANDLER_ARGS_PASSTHRU);
 }
 
-static void xdebug_count_line(char *filename, int lineno, int executable, int deadcode)
+static void xdebug_count_line(zend_string *filename, int lineno, int executable, int deadcode)
 {
 	xdebug_coverage_file *file;
 	xdebug_coverage_line *line;
 
-	if (XG_COV(previous_filename) && strcmp(XG_COV(previous_filename), filename) == 0) {
+	if (XG_COV(previous_filename) && zend_string_equals(XG_COV(previous_filename), filename)) {
 		file = XG_COV(previous_file);
 	} else {
 		/* Check if the file already exists in the hash */
-		if (!xdebug_hash_find(XG_COV(code_coverage_info), filename, strlen(filename), (void *) &file)) {
+		if (!xdebug_hash_find(XG_COV(code_coverage_info), ZSTR_VAL(filename), ZSTR_LEN(filename), (void *) &file)) {
 			/* The file does not exist, so we add it to the hash */
 			file = xdebug_coverage_file_ctor(filename);
 
-			xdebug_hash_add(XG_COV(code_coverage_info), filename, strlen(filename), file);
+			xdebug_hash_add(XG_COV(code_coverage_info), ZSTR_VAL(filename), ZSTR_LEN(filename), file);
 		}
-		XG_COV(previous_filename) = file->name;
+		if (XG_COV(previous_filename)) {
+			zend_string_release(XG_COV(previous_filename));
+		}
+		XG_COV(previous_filename) = zend_string_copy(file->name);
 		XG_COV(previous_file) = file;
 	}
 
@@ -172,26 +219,34 @@ static void xdebug_count_line(char *filename, int lineno, int executable, int de
 	}
 }
 
-int xdebug_common_override_handler(XDEBUG_OPCODE_HANDLER_ARGS)
+static int xdebug_common_override_handler(XDEBUG_OPCODE_HANDLER_ARGS)
 {
 	zend_op_array *op_array = &execute_data->func->op_array;
 	const zend_op *cur_opcode = execute_data->opline;
 
 	if (!op_array->reserved[XG_COV(code_coverage_filter_offset)] && XG_COV(code_coverage_active)) {
 		int      lineno;
-		char    *file;
 
 		lineno = cur_opcode->lineno;
-		file = (char*) STR_NAME_VAL(op_array->filename);
 
-		xdebug_print_opcode_info('C', execute_data, cur_opcode);
-		xdebug_count_line(file, lineno, 0, 0);
+		xdebug_print_opcode_info(execute_data, cur_opcode);
+		xdebug_count_line(op_array->filename, lineno, 0, 0);
 	}
 
 	return xdebug_call_original_opcode_handler_if_set(cur_opcode->opcode, XDEBUG_OPCODE_HANDLER_ARGS_PASSTHRU);
 }
 
-static void prefill_from_opcode(char *fn, zend_op opcode, int deadcode)
+static int xdebug_coverage_include_or_eval_handler(XDEBUG_OPCODE_HANDLER_ARGS)
+{
+	zend_op_array *op_array = &execute_data->func->op_array;
+	const zend_op *opline = execute_data->opline;
+
+	xdebug_coverage_record_if_active(execute_data, op_array);
+
+	return xdebug_call_original_opcode_handler_if_set(opline->opcode, XDEBUG_OPCODE_HANDLER_ARGS_PASSTHRU);
+}
+
+static void prefill_from_opcode(zend_string *filename, zend_op opcode, int deadcode)
 {
 	if (
 		opcode.opcode != ZEND_NOP &&
@@ -207,7 +262,7 @@ static void prefill_from_opcode(char *fn, zend_op opcode, int deadcode)
 		&& opcode.opcode != ZEND_FAST_CALL
 		&& opcode.opcode != ZEND_RECV_VARIADIC
 	) {
-		xdebug_count_line(fn, opcode.lineno, 1, deadcode);
+		xdebug_count_line(filename, opcode.lineno, 1, deadcode);
 	}
 }
 
@@ -224,6 +279,7 @@ static int xdebug_find_jumps(zend_op_array *opa, unsigned int position, size_t *
 		jumps[0] = XDEBUG_ZNODE_JMP_LINE(opcode.op1, position, base_address);
 		*jump_count = 1;
 		return 1;
+
 	} else if (
 		opcode.opcode == ZEND_JMPZ ||
 		opcode.opcode == ZEND_JMPNZ ||
@@ -234,21 +290,25 @@ static int xdebug_find_jumps(zend_op_array *opa, unsigned int position, size_t *
 		jumps[1] = XDEBUG_ZNODE_JMP_LINE(opcode.op2, position, base_address);
 		*jump_count = 2;
 		return 1;
+
 	} else if (opcode.opcode == ZEND_JMPZNZ) {
 		jumps[0] = XDEBUG_ZNODE_JMP_LINE(opcode.op2, position, base_address);
 		jumps[1] = position + ((int32_t) opcode.extended_value / (int32_t) sizeof(zend_op));
 		*jump_count = 2;
 		return 1;
+
 	} else if (opcode.opcode == ZEND_FE_FETCH_R || opcode.opcode == ZEND_FE_FETCH_RW) {
 		jumps[0] = position + 1;
 		jumps[1] = position + (opcode.extended_value / sizeof(zend_op));
 		*jump_count = 2;
 		return 1;
+
 	} else if (opcode.opcode == ZEND_FE_RESET_R || opcode.opcode == ZEND_FE_RESET_RW) {
 		jumps[0] = position + 1;
 		jumps[1] = XDEBUG_ZNODE_JMP_LINE(opcode.op2, position, base_address);
 		*jump_count = 2;
 		return 1;
+
 	} else if (opcode.opcode == ZEND_CATCH) {
 		*jump_count = 2;
 		jumps[0] = position + 1;
@@ -267,6 +327,7 @@ static int xdebug_find_jumps(zend_op_array *opa, unsigned int position, size_t *
 			jumps[1] = XDEBUG_JMP_EXIT;
 		}
 		return 1;
+
 	} else if (opcode.opcode == ZEND_GOTO) {
 		jumps[0] = XDEBUG_ZNODE_JMP_LINE(opcode.op1, position, base_address);
 		*jump_count = 1;
@@ -277,21 +338,28 @@ static int xdebug_find_jumps(zend_op_array *opa, unsigned int position, size_t *
 		jumps[1] = position + 1;
 		*jump_count = 2;
 		return 1;
+
 	} else if (opcode.opcode == ZEND_FAST_RET) {
 		jumps[0] = XDEBUG_JMP_EXIT;
 		*jump_count = 1;
 		return 1;
+
 	} else if (
 		opcode.opcode == ZEND_GENERATOR_RETURN ||
 		opcode.opcode == ZEND_EXIT ||
 		opcode.opcode == ZEND_THROW ||
+#if PHP_VERSION_ID >= 80000
+		opcode.opcode == ZEND_MATCH_ERROR ||
+#endif
 		opcode.opcode == ZEND_RETURN
 	) {
 		jumps[0] = XDEBUG_JMP_EXIT;
 		*jump_count = 1;
 		return 1;
-#if PHP_VERSION_ID >= 70200
 	} else if (
+# if PHP_VERSION_ID >= 80000
+		opcode.opcode == ZEND_MATCH ||
+# endif
 		opcode.opcode == ZEND_SWITCH_LONG ||
 		opcode.opcode == ZEND_SWITCH_STRING
 	) {
@@ -318,12 +386,17 @@ static int xdebug_find_jumps(zend_op_array *opa, unsigned int position, size_t *
 		jumps[*jump_count] = position + (opcode.extended_value / sizeof(zend_op));
 		(*jump_count)++;
 
-		/* The 'next' opcode */
-		jumps[*jump_count] = position + 1;
-		(*jump_count)++;
+# if PHP_VERSION_ID >= 80000
+		if (opcode.opcode != ZEND_MATCH) {
+# endif
+			/* The 'next' opcode */
+			jumps[*jump_count] = position + 1;
+			(*jump_count)++;
+# if PHP_VERSION_ID >= 80000
+		}
+# endif
 
 		return 1;
-#endif
 	}
 
 	return 0;
@@ -428,7 +501,7 @@ static void xdebug_analyse_oparray(zend_op_array *opa, xdebug_set *set, xdebug_b
 	}
 }
 
-static void prefill_from_oparray(char *filename, zend_op_array *op_array)
+static void prefill_from_oparray(zend_string *filename, zend_op_array *op_array)
 {
 	unsigned int i;
 	xdebug_set *set = NULL;
@@ -477,14 +550,14 @@ static void prefill_from_oparray(char *filename, zend_op_array *op_array)
 		xdebug_set_free(set);
 	}
 	if (branch_info) {
-		char *function_name;
+		char function_name[1024];
 		xdebug_func func_info;
 
 		xdebug_build_fname_from_oparray(&func_info, op_array);
-		function_name = xdebug_func_format(&func_info);
+		xdebug_func_format(function_name, sizeof(function_name), &func_info);
 
-		if (func_info.class) {
-			xdfree(func_info.class);
+		if (func_info.class_name) {
+			zend_string_release(func_info.class_name);
 		}
 		if (func_info.function) {
 			xdfree(func_info.function);
@@ -493,8 +566,6 @@ static void prefill_from_oparray(char *filename, zend_op_array *op_array)
 		xdebug_branch_post_process(op_array, branch_info);
 		xdebug_branch_find_paths(branch_info);
 		xdebug_branch_info_add_branches_and_paths(filename, (char*) function_name, branch_info);
-
-		xdfree(function_name);
 	}
 }
 
@@ -502,7 +573,7 @@ static int prefill_from_function_table(zend_op_array *opa)
 {
 	if (opa->type == ZEND_USER_FUNCTION) {
 		if ((long) opa->reserved[XG_COV(dead_code_analysis_tracker_offset)] < XG_COV(dead_code_last_start_id)) {
-			prefill_from_oparray((char*) STR_NAME_VAL(opa->filename), opa);
+			prefill_from_oparray(opa->filename, opa);
 		}
 	}
 
@@ -537,7 +608,7 @@ static void xdebug_prefill_code_coverage(zend_op_array *op_array)
 	zend_class_entry *class_entry;
 
 	if ((long) op_array->reserved[XG_COV(dead_code_analysis_tracker_offset)] < XG_COV(dead_code_last_start_id)) {
-		prefill_from_oparray((char*) STR_NAME_VAL(op_array->filename), op_array);
+		prefill_from_oparray(op_array->filename, op_array);
 	}
 
 	ZEND_HASH_REVERSE_FOREACH_PTR(CG(function_table), function_op_array) {
@@ -572,7 +643,7 @@ void xdebug_code_coverage_start_of_function(zend_op_array *op_array, char *funct
 	XG_COV(branches).last_branch_nr[XG_BASE(level)] = -1;
 }
 
-void xdebug_code_coverage_end_of_function(zend_op_array *op_array, char *file_name, char *function_name)
+void xdebug_code_coverage_end_of_function(zend_op_array *op_array, zend_string *filename, char *function_name)
 {
 	xdebug_str str = XDEBUG_STR_INITIALIZER;
 	xdebug_path *path = xdebug_path_info_get_path_for_level(XG_COV(paths_stack), XG_BASE(level));
@@ -583,7 +654,7 @@ void xdebug_code_coverage_end_of_function(zend_op_array *op_array, char *file_na
 
 	xdebug_create_key_for_path(path, &str);
 
-	xdebug_branch_info_mark_end_of_function_reached(file_name, function_name, str.d, str.l);
+	xdebug_branch_info_mark_end_of_function_reached(filename, function_name, str.d, str.l);
 
 	xdfree(str.d);
 
@@ -596,53 +667,66 @@ PHP_FUNCTION(xdebug_start_code_coverage)
 {
 	zend_long options = 0;
 
+	if (!XDEBUG_MODE_IS(XDEBUG_MODE_COVERAGE)) {
+		php_error(E_WARNING, "Code coverage needs to be enabled in php.ini by setting 'xdebug.mode' to 'coverage'");
+		RETURN_FALSE;
+	}
+
 	if (zend_parse_parameters(ZEND_NUM_ARGS(), "|l", &options) == FAILURE) {
 		return;
 	}
+
 	XG_COV(code_coverage_unused) = (options & XDEBUG_CC_OPTION_UNUSED);
 	XG_COV(code_coverage_dead_code_analysis) = (options & XDEBUG_CC_OPTION_DEAD_CODE);
 	XG_COV(code_coverage_branch_check) = (options & XDEBUG_CC_OPTION_BRANCH_CHECK);
 
-	if (!XINI_COV(enable)) {
-		php_error(E_WARNING, "Code coverage needs to be enabled in php.ini by setting 'xdebug.coverage_enable' to '1'.");
-		RETURN_FALSE;
-	} else {
-		XG_COV(code_coverage_active) = 1;
-		RETURN_TRUE;
-	}
+	XG_COV(code_coverage_active) = 1;
+	RETURN_TRUE;
 }
 
 PHP_FUNCTION(xdebug_stop_code_coverage)
 {
 	zend_long cleanup = 1;
 
+	if (!XDEBUG_MODE_IS(XDEBUG_MODE_COVERAGE)) {
+		RETURN_FALSE;
+	}
+
 	if (zend_parse_parameters(ZEND_NUM_ARGS(), "|l", &cleanup) == FAILURE) {
 		return;
 	}
-	if (XG_COV(code_coverage_active)) {
-		if (cleanup) {
-			XG_COV(previous_filename) = NULL;
-			XG_COV(previous_file) = NULL;
-			XG_COV(previous_mark_filename) = NULL;
-			XG_COV(previous_mark_file) = NULL;
-			xdebug_hash_destroy(XG_COV(code_coverage_info));
-			XG_COV(code_coverage_info) = xdebug_hash_alloc(32, xdebug_coverage_file_dtor);
-			XG_COV(dead_code_last_start_id)++;
-			xdebug_path_info_dtor(XG_COV(paths_stack));
-			XG_COV(paths_stack) = xdebug_path_info_ctor();
-		}
-		XG_COV(code_coverage_active) = 0;
-		RETURN_TRUE;
+	if (!XG_COV(code_coverage_active)) {
+		RETURN_FALSE;
 	}
-	RETURN_FALSE;
+
+	if (cleanup) {
+		zend_string_release(XG_COV(previous_filename));
+		XG_COV(previous_filename) = NULL;
+		XG_COV(previous_file) = NULL;
+		zend_string_release(XG_COV(previous_mark_filename));
+		XG_COV(previous_mark_filename) = NULL;
+		XG_COV(previous_mark_file) = NULL;
+		xdebug_hash_destroy(XG_COV(code_coverage_info));
+		XG_COV(code_coverage_info) = xdebug_hash_alloc(32, xdebug_coverage_file_dtor);
+		XG_COV(dead_code_last_start_id)++;
+		xdebug_path_info_dtor(XG_COV(paths_stack));
+		XG_COV(paths_stack) = xdebug_path_info_ctor();
+	}
+
+	XG_COV(code_coverage_active) = 0;
+
+	RETURN_TRUE;
 }
 
-
+#if PHP_VERSION_ID >= 80000
+static int xdebug_lineno_cmp(Bucket *f, Bucket *s)
+{
+#else
 static int xdebug_lineno_cmp(const void *a, const void *b)
 {
 	Bucket *f = (Bucket *) a;
 	Bucket *s = (Bucket *) b;
-
+#endif
 	if (f->h < s->h) {
 		return -1;
 	} else if (f->h > s->h) {
@@ -786,7 +870,7 @@ static void add_file(void *ret, xdebug_hash_element *e)
 	zend_hash_sort(target_hash, xdebug_lineno_cmp, 0);
 
 	/* Add the branch and path info */
-	if (file->has_branch_info) {
+	if (XG_COV(code_coverage_branch_check)) {
 		XDEBUG_MAKE_STD_ZVAL(file_info);
 		array_init(file_info);
 
@@ -798,11 +882,11 @@ static void add_file(void *ret, xdebug_hash_element *e)
 		add_assoc_zval_ex(file_info, "lines", HASH_KEY_SIZEOF("lines"), lines);
 		add_assoc_zval_ex(file_info, "functions", HASH_KEY_SIZEOF("functions"), functions);
 
-		add_assoc_zval_ex(retval, file->name, HASH_KEY_STRLEN(file->name), file_info);
+		add_assoc_zval_ex(retval, ZSTR_VAL(file->name), ZSTR_LEN(file->name), file_info);
 		efree(functions);
 		efree(file_info);
 	} else {
-		add_assoc_zval_ex(retval, file->name, HASH_KEY_STRLEN(file->name), lines);
+		add_assoc_zval_ex(retval, ZSTR_VAL(file->name), ZSTR_LEN(file->name), lines);
 	}
 
 	efree(lines);
@@ -811,9 +895,12 @@ static void add_file(void *ret, xdebug_hash_element *e)
 PHP_FUNCTION(xdebug_get_code_coverage)
 {
 	array_init(return_value);
-	if (XG_COV(code_coverage_info)) {
-		xdebug_hash_apply(XG_COV(code_coverage_info), (void *) return_value, add_file);
+
+	if (!XG_COV(code_coverage_info)) {
+		return;
 	}
+
+	xdebug_hash_apply(XG_COV(code_coverage_info), (void *) return_value, add_file);
 }
 
 PHP_FUNCTION(xdebug_get_function_count)
@@ -843,48 +930,24 @@ void xdebug_init_coverage_globals(xdebug_coverage_globals_t *xg)
 	xg->code_coverage_filter_offset = zend_xdebug_filter_offset;
 }
 
-void xdebug_coverage_count_line_if_active(zend_op_array *op_array, char *file, int lineno)
+void xdebug_coverage_count_line_if_active(zend_op_array *op_array, zend_string *file, int lineno)
 {
 	if (XG_COV(code_coverage_active) && !op_array->reserved[XG_COV(code_coverage_filter_offset)]) {
 		xdebug_count_line(file, lineno, 0, 0);
 	}
 }
 
-void xdebug_coverage_count_line_if_branch_check_active(zend_op_array *op_array, char *file, int lineno)
+void xdebug_coverage_count_line_if_branch_check_active(zend_op_array *op_array, zend_string *file, int lineno)
 {
 	if (XG_COV(code_coverage_active) && XG_COV(code_coverage_branch_check)) {
 		xdebug_coverage_count_line_if_active(op_array, file, lineno);
 	}
 }
 
-void xdebug_coverage_record_assign_if_active(zend_execute_data *execute_data, zend_op_array *op_array, int do_cc)
+void xdebug_coverage_record_if_active(zend_execute_data *execute_data, zend_op_array *op_array)
 {
 	if (!op_array->reserved[XG_COV(code_coverage_filter_offset)] && XG_COV(code_coverage_active)) {
-		const zend_op *cur_opcode = execute_data->opline;
-		xdebug_print_opcode_info('=', execute_data, cur_opcode);
-
-		if (do_cc) {
-			char *file   = (char*) STR_NAME_VAL(op_array->filename);
-			int   lineno = cur_opcode->lineno;
-
-			xdebug_count_line(file, lineno, 0, 0);
-		}
-	}
-}
-
-void xdebug_coverage_record_include_if_active(zend_execute_data *execute_data, zend_op_array *op_array)
-{
-	if (!op_array->reserved[XG_COV(code_coverage_filter_offset)] && XG_COV(code_coverage_active)) {
-		const zend_op *cur_opcode = execute_data->opline;
-		xdebug_print_opcode_info('I', execute_data, cur_opcode);
-	}
-}
-
-void xdebug_coverage_record_silence_if_active(zend_execute_data *execute_data, zend_op_array *op_array)
-{
-	if (!op_array->reserved[XG_COV(code_coverage_filter_offset)] && XG_COV(code_coverage_active)) {
-		const zend_op *cur_opcode = execute_data->opline;
-		xdebug_print_opcode_info('S', execute_data, cur_opcode);
+		xdebug_print_opcode_info(execute_data, execute_data->opline);
 	}
 }
 
@@ -895,18 +958,21 @@ void xdebug_coverage_compile_file(zend_op_array *op_array)
 	}
 }
 
-int xdebug_coverage_execute_ex(function_stack_entry *fse, zend_op_array *op_array, char **tmp_file_name, char **tmp_function_name)
+int xdebug_coverage_execute_ex(function_stack_entry *fse, zend_op_array *op_array, zend_string **tmp_filename, char **tmp_function_name)
 {
 	xdebug_func func_info;
 
 	if (!fse->filtered_code_coverage && XG_COV(code_coverage_active) && XG_COV(code_coverage_unused)) {
-		*tmp_file_name = xdstrdup(STR_NAME_VAL(op_array->filename));
+		char buffer[1024];
+
+		*tmp_filename = zend_string_copy(op_array->filename);
 		xdebug_build_fname_from_oparray(&func_info, op_array);
-		*tmp_function_name = xdebug_func_format(&func_info);
+		xdebug_func_format(buffer, sizeof(buffer), &func_info);
+		*tmp_function_name = xdstrdup(buffer);
 		xdebug_code_coverage_start_of_function(op_array, *tmp_function_name);
 
-		if (func_info.class) {
-			xdfree(func_info.class);
+		if (func_info.class_name) {
+			zend_string_release(func_info.class_name);
 		}
 		if (func_info.function) {
 			xdfree(func_info.function);
@@ -917,152 +983,147 @@ int xdebug_coverage_execute_ex(function_stack_entry *fse, zend_op_array *op_arra
 	return 0;
 }
 
-void xdebug_coverage_execute_ex_end(function_stack_entry *fse, zend_op_array *op_array, char *tmp_file_name, char *tmp_function_name)
+void xdebug_coverage_execute_ex_end(function_stack_entry *fse, zend_op_array *op_array, zend_string *tmp_filename, char *tmp_function_name)
 {
 	/* Check which path has been used */
 	if (!fse->filtered_code_coverage && XG_COV(code_coverage_active) && XG_COV(code_coverage_unused)) {
-		xdebug_code_coverage_end_of_function(op_array, tmp_file_name, tmp_function_name);
-		xdfree(tmp_function_name);
-		xdfree(tmp_file_name);
+		xdebug_code_coverage_end_of_function(op_array, tmp_filename, tmp_function_name);
 	}
+	xdfree(tmp_function_name);
+	zend_string_release(tmp_filename);
 }
 
 void xdebug_coverage_init_oparray(zend_op_array *op_array)
 {
-	op_array->reserved[XG_COV(dead_code_analysis_tracker_offset)] = 0;
+	function_stack_entry tmp_fse;
 
-	if (XG_BASE(filter_type_code_coverage) != XDEBUG_FILTER_NONE) {
-		function_stack_entry tmp_fse;
-
-		tmp_fse.filename = STR_NAME_VAL(op_array->filename);
-		xdebug_build_fname_from_oparray(&tmp_fse.function, op_array);
-		xdebug_filter_run_internal(&tmp_fse, XDEBUG_FILTER_CODE_COVERAGE, &tmp_fse.filtered_code_coverage, XG_BASE(filter_type_code_coverage), XG_BASE(filters_code_coverage));
-		xdebug_func_dtor_by_ref(&tmp_fse.function);
-
-		op_array->reserved[XG_COV(code_coverage_filter_offset)] = (void*) tmp_fse.filtered_code_coverage;
+	if (XG_BASE(filter_type_code_coverage) == XDEBUG_FILTER_NONE) {
+		op_array->reserved[XG_COV(dead_code_analysis_tracker_offset)] = 0;
+		return;
 	}
+
+	tmp_fse.filename = op_array->filename;
+	xdebug_build_fname_from_oparray(&tmp_fse.function, op_array);
+	xdebug_filter_run_internal(&tmp_fse, XDEBUG_FILTER_CODE_COVERAGE, &tmp_fse.filtered_code_coverage, XG_BASE(filter_type_code_coverage), XG_BASE(filters_code_coverage));
+	xdebug_func_dtor_by_ref(&tmp_fse.function);
+
+	op_array->reserved[XG_COV(code_coverage_filter_offset)] = (void*) (size_t) tmp_fse.filtered_code_coverage;
 }
 
-#if PHP_VERSION_ID >= 70200
 static int xdebug_switch_handler(XDEBUG_OPCODE_HANDLER_ARGS)
 {
 	const zend_op *cur_opcode = execute_data->opline;
 
-	if (XG_COV(code_coverage_active)) {
-		execute_data->opline++;
-		return ZEND_USER_OPCODE_CONTINUE;
+	if (!XG_COV(code_coverage_active)) {
+		return xdebug_call_original_opcode_handler_if_set(cur_opcode->opcode, XDEBUG_OPCODE_HANDLER_ARGS_PASSTHRU);
 	}
 
-	return xdebug_call_original_opcode_handler_if_set(cur_opcode->opcode, XDEBUG_OPCODE_HANDLER_ARGS_PASSTHRU);
+	execute_data->opline++;
+	return ZEND_USER_OPCODE_CONTINUE;
 }
-#endif
-
-#define XDEBUG_OPCODE_OVERRIDE(f) \
-	int xdebug_##f##_handler(zend_execute_data *execute_data) \
-	{ \
-		return xdebug_common_override_handler(execute_data); \
-	}
 
 
 void xdebug_coverage_minit(INIT_FUNC_ARGS)
 {
-	zend_extension dummy_ext;
+	int i;
 
 	/* Get reserved offsets */
+#if PHP_VERSION_ID >= 80000
+	zend_xdebug_cc_run_offset = zend_get_resource_handle(XDEBUG_NAME);
+	zend_xdebug_filter_offset = zend_get_resource_handle(XDEBUG_NAME);
+#else
+	zend_extension dummy_ext;
 	zend_xdebug_cc_run_offset = zend_get_resource_handle(&dummy_ext);
 	zend_xdebug_filter_offset = zend_get_resource_handle(&dummy_ext);
+#endif
+
+	xdebug_register_with_opcode_multi_handler(ZEND_ASSIGN, xdebug_common_override_handler);
+	xdebug_register_with_opcode_multi_handler(ZEND_QM_ASSIGN, xdebug_common_override_handler);
+	xdebug_register_with_opcode_multi_handler(ZEND_INCLUDE_OR_EVAL, xdebug_coverage_include_or_eval_handler);
 
 	/* Overload opcodes for code coverage */
-	if (XINI_COV(enable)) {
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_JMP);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_JMPZ);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_JMPZ_EX);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_JMPNZ);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_IS_IDENTICAL);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_IS_NOT_IDENTICAL);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_IS_EQUAL);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_IS_NOT_EQUAL);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_IS_SMALLER);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_IS_SMALLER_OR_EQUAL);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_BOOL_NOT);
+	xdebug_set_opcode_handler(ZEND_JMP, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_JMPZ, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_JMPZ_EX, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_JMPNZ, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_IS_IDENTICAL, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_IS_NOT_IDENTICAL, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_IS_EQUAL, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_IS_NOT_EQUAL, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_IS_SMALLER, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_IS_SMALLER_OR_EQUAL, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_BOOL_NOT, xdebug_common_override_handler);
 
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_ADD);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_SUB);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_MUL);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_DIV);
+	xdebug_set_opcode_handler(ZEND_ADD, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_SUB, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_MUL, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_DIV, xdebug_common_override_handler);
 
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_ADD_ARRAY_ELEMENT);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_RETURN);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_RETURN_BY_REF);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_EXT_STMT);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_SEND_VAR);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_SEND_VAR_NO_REF);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_SEND_VAR_NO_REF_EX);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_SEND_REF);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_SEND_VAL);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_SEND_VAL_EX);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_SEND_VAR_EX);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_NEW);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_EXT_FCALL_BEGIN);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_INIT_METHOD_CALL);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_INIT_STATIC_METHOD_CALL);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_INIT_FCALL);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_INIT_NS_FCALL_BY_NAME);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_CATCH);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_BOOL);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_INIT_ARRAY);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_FETCH_DIM_R);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_FETCH_DIM_W);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_FETCH_OBJ_R);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_FETCH_OBJ_W);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_FETCH_OBJ_FUNC_ARG);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_FETCH_DIM_FUNC_ARG);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_FETCH_STATIC_PROP_FUNC_ARG);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_FETCH_DIM_UNSET);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_FETCH_OBJ_UNSET);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_FETCH_CLASS);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_FETCH_CONSTANT);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_FETCH_CLASS_CONSTANT);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_CONCAT);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_ISSET_ISEMPTY_DIM_OBJ);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_ISSET_ISEMPTY_PROP_OBJ);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_CASE);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_DECLARE_LAMBDA_FUNCTION);
+	xdebug_set_opcode_handler(ZEND_ADD_ARRAY_ELEMENT, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_RETURN, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_RETURN_BY_REF, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_EXT_STMT, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_SEND_VAR, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_SEND_VAR_NO_REF, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_SEND_VAR_NO_REF_EX, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_SEND_REF, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_SEND_VAL, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_SEND_VAL_EX, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_SEND_VAR_EX, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_NEW, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_EXT_FCALL_BEGIN, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_INIT_METHOD_CALL, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_INIT_STATIC_METHOD_CALL, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_INIT_FCALL, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_INIT_NS_FCALL_BY_NAME, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_CATCH, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_BOOL, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_INIT_ARRAY, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_FETCH_DIM_R, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_FETCH_DIM_W, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_FETCH_OBJ_R, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_FETCH_OBJ_W, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_FETCH_OBJ_FUNC_ARG, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_FETCH_DIM_FUNC_ARG, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_FETCH_STATIC_PROP_FUNC_ARG, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_FETCH_DIM_UNSET, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_FETCH_OBJ_UNSET, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_FETCH_CLASS, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_FETCH_CONSTANT, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_FETCH_CLASS_CONSTANT, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_CONCAT, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_ISSET_ISEMPTY_DIM_OBJ, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_ISSET_ISEMPTY_PROP_OBJ, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_CASE, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_DECLARE_LAMBDA_FUNCTION, xdebug_common_override_handler);
 #if PHP_VERSION_ID < 70400
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_ADD_TRAIT);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_BIND_TRAITS);
+	xdebug_set_opcode_handler(ZEND_ADD_TRAIT, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_BIND_TRAITS, xdebug_common_override_handler);
 #endif
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_INSTANCEOF);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_FAST_RET);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_ROPE_ADD);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_ROPE_END);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_COALESCE);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_TYPE_CHECK);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_GENERATOR_CREATE);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_BIND_STATIC);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_BIND_LEXICAL);
+	xdebug_set_opcode_handler(ZEND_INSTANCEOF, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_FAST_RET, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_ROPE_ADD, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_ROPE_END, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_COALESCE, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_TYPE_CHECK, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_GENERATOR_CREATE, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_BIND_STATIC, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_BIND_LEXICAL, xdebug_common_override_handler);
 #if PHP_VERSION_ID >= 70400
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_DECLARE_CLASS);
-		XDEBUG_SET_OPCODE_OVERRIDE_COMMON(ZEND_DECLARE_CLASS_DELAYED);
+	xdebug_set_opcode_handler(ZEND_DECLARE_CLASS, xdebug_common_override_handler);
+	xdebug_set_opcode_handler(ZEND_DECLARE_CLASS_DELAYED, xdebug_common_override_handler);
 #endif
-#if PHP_VERSION_ID >= 70200
-		xdebug_set_opcode_handler(ZEND_SWITCH_STRING, xdebug_switch_handler);
-		xdebug_set_opcode_handler(ZEND_SWITCH_LONG, xdebug_switch_handler);
-#endif
-	}
+	xdebug_set_opcode_handler(ZEND_SWITCH_STRING, xdebug_switch_handler);
+	xdebug_set_opcode_handler(ZEND_SWITCH_LONG, xdebug_switch_handler);
 
 	/* Override all the other opcodes so that we can mark when we hit a branch
 	 * start one */
-	if (XINI_COV(enable)) {
-		int i;
-
-		for (i = 0; i < 256; i++) {
-			if (i == ZEND_HANDLE_EXCEPTION) {
-				continue;
-			}
-			if (!xdebug_isset_opcode_handler(i)) {
-				xdebug_set_opcode_handler(i, xdebug_check_branch_entry_handler);
-			}
+	for (i = 0; i < 256; i++) {
+		if (i == ZEND_HANDLE_EXCEPTION) {
+			continue;
+		}
+		if (!xdebug_isset_opcode_handler(i)) {
+			xdebug_set_opcode_handler(i, xdebug_check_branch_entry_handler);
 		}
 	}
 
@@ -1111,5 +1172,13 @@ void xdebug_coverage_post_deactivate(void)
 		XG_COV(branches).last_branch_nr = NULL;
 		XG_COV(branches).size = 0;
 	}
-	XG_COV(previous_mark_filename) = NULL;
+
+	if (XG_COV(previous_filename)) {
+		zend_string_release(XG_COV(previous_filename));
+		XG_COV(previous_filename) = NULL;
+	}
+	if (XG_COV(previous_mark_filename)) {
+		zend_string_release(XG_COV(previous_mark_filename));
+		XG_COV(previous_mark_filename) = NULL;
+	}
 }
