@@ -17,6 +17,7 @@
  */
 
 #include "var_export_html.h"
+#include "lib_private.h"
 
 ZEND_EXTERN_MODULE_GLOBALS(xdebug)
 
@@ -44,22 +45,22 @@ static int xdebug_array_element_export_html(zval *zv_nptr, zend_ulong index_key,
 	if (options->runtime[level].current_element_nr >= options->runtime[level].start_element_nr &&
 		options->runtime[level].current_element_nr < options->runtime[level].end_element_nr)
 	{
-		xdebug_str_add(str, xdebug_sprintf("%*s", (level * 4) - 2, ""), 1);
+		xdebug_str_add_fmt(str, "%*s", (level * 4) - 2, "");
 
 		if (HASH_KEY_IS_NUMERIC(hash_key)) { /* numeric key */
-			xdebug_str_add(str, xdebug_sprintf(XDEBUG_INT_FMT " <font color='%s'>=&gt;</font> ", index_key, COLOR_POINTER), 1);
+			xdebug_str_add_fmt(str, XDEBUG_INT_FMT " <font color='%s'>=&gt;</font> ", index_key, COLOR_POINTER);
 		} else { /* string key */
-			xdebug_str_addl(str, "'", 1, 0);
+			xdebug_str_addc(str, '\'');
 			tmp_str = xdebug_xmlize((char*) HASH_APPLY_KEY_VAL(hash_key), HASH_APPLY_KEY_LEN(hash_key) - 1, &newlen);
 			xdebug_str_addl(str, tmp_str, newlen, 0);
 			efree(tmp_str);
-			xdebug_str_add(str, xdebug_sprintf("' <font color='%s'>=&gt;</font> ", COLOR_POINTER), 1);
+			xdebug_str_add_fmt(str, "' <font color='%s'>=&gt;</font> ", COLOR_POINTER);
 		}
 		xdebug_var_export_html(zv, str, level + 1, debug_zval, options);
 	}
 	if (options->runtime[level].current_element_nr == options->runtime[level].end_element_nr) {
-		xdebug_str_add(str, xdebug_sprintf("%*s", (level * 4) - 2, ""), 1);
-		xdebug_str_addl(str, "<i>more elements...</i>\n", 24, 0);
+		xdebug_str_add_fmt(str, "%*s", (level * 4) - 2, "");
+		xdebug_str_add_literal(str, "<i>more elements...</i>\n");
 	}
 	options->runtime[level].current_element_nr++;
 	return 0;
@@ -72,7 +73,7 @@ static int xdebug_object_element_export_html(zval *object, zval *zv_nptr, zend_u
 	if (options->runtime[level].current_element_nr >= options->runtime[level].start_element_nr &&
 		options->runtime[level].current_element_nr < options->runtime[level].end_element_nr)
 	{
-		xdebug_str_add(str, xdebug_sprintf("%*s", (level * 4) - 2, ""), 1);
+		xdebug_str_add_fmt(str, "%*s", (level * 4) - 2, "");
 
 		if (!HASH_KEY_IS_NUMERIC(hash_key)) {
 			char       *prop_class_name;
@@ -85,16 +86,16 @@ static int xdebug_object_element_export_html(zval *object, zval *zv_nptr, zend_u
 #endif
 			property_name = xdebug_get_property_info((char*) HASH_APPLY_KEY_VAL(hash_key), HASH_APPLY_KEY_LEN(hash_key), &modifier, &prop_class_name);
 
-			xdebug_str_add(str, xdebug_sprintf("<i>%s</i> ", modifier), 1);
+			xdebug_str_add_fmt(str, "<i>%s</i> ", modifier);
 			if (property_type) {
-				xdebug_str_add(str, xdebug_sprintf("<i>%s</i> ", property_type->d), 1);
+				xdebug_str_add_fmt(str, "<i>%s</i> ", property_type->d);
 			}
 			xdebug_str_addc(str, '\'');
 			xdebug_str_add_str(str, property_name);
 			if (strcmp(modifier, "private") != 0 || strcmp(class_name, prop_class_name) == 0) {
-				xdebug_str_add(str, xdebug_sprintf("' <font color='%s'>=&gt;</font> ", COLOR_POINTER), 1);
+				xdebug_str_add_fmt(str, "' <font color='%s'>=&gt;</font> ", COLOR_POINTER);
 			} else {
-				xdebug_str_add(str, xdebug_sprintf("' <small>(%s)</small> <font color='%s'>=&gt;</font> ", prop_class_name, COLOR_POINTER), 1);
+				xdebug_str_add_fmt(str, "' <small>(%s)</small> <font color='%s'>=&gt;</font> ", prop_class_name, COLOR_POINTER);
 			}
 
 			if (property_type) {
@@ -103,13 +104,13 @@ static int xdebug_object_element_export_html(zval *object, zval *zv_nptr, zend_u
 			xdebug_str_free(property_name);
 			xdfree(prop_class_name);
 		} else {
-			xdebug_str_add(str, xdebug_sprintf("<i>public</i> " XDEBUG_INT_FMT " <font color='%s'>=&gt;</font> ", index_key, COLOR_POINTER), 1);
+			xdebug_str_add_fmt(str, "<i>public</i> " XDEBUG_INT_FMT " <font color='%s'>=&gt;</font> ", index_key, COLOR_POINTER);
 		}
 		xdebug_var_export_html(zv, str, level + 1, debug_zval, options);
 	}
 	if (options->runtime[level].current_element_nr == options->runtime[level].end_element_nr) {
-		xdebug_str_add(str, xdebug_sprintf("%*s", (level * 4) - 2, ""), 1);
-		xdebug_str_addl(str, "<i>more elements...</i>\n", 24, 0);
+		xdebug_str_add_fmt(str, "%*s", (level * 4) - 2, "");
+		xdebug_str_add_literal(str, "<i>more elements...</i>\n");
 	}
 	options->runtime[level].current_element_nr++;
 	return 0;
@@ -127,59 +128,67 @@ void xdebug_var_export_html(zval **struc, xdebug_str *str, int level, int debug_
 	zend_string *key;
 	zval *val;
 	zval *tmpz;
+	int   z_type;
+
+	z_type = Z_TYPE_P(*struc);
 
 	if (debug_zval) {
 		xdebug_add_variable_attributes(str, *struc, XDEBUG_VAR_ATTR_HTML);
 	}
-	if (Z_TYPE_P(*struc) == IS_INDIRECT) {
+	if (z_type == IS_INDIRECT) {
 		tmpz = Z_INDIRECT_P(*struc);
 		struc = &tmpz;
+		z_type = Z_TYPE_P(*struc);
 	}
-	if (Z_TYPE_P(*struc) == IS_REFERENCE) {
+	if (z_type == IS_REFERENCE) {
 		tmpz = &((*struc)->value.ref->val);
 		struc = &tmpz;
+		z_type = Z_TYPE_P(*struc);
 	}
 
-	switch (Z_TYPE_P(*struc)) {
+	switch (z_type) {
 		case IS_TRUE:
+			xdebug_str_add_fmt(str, "<small>boolean</small> <font color='%s'>true</font>", COLOR_BOOL);
+			break;
+
 		case IS_FALSE:
-			xdebug_str_add(str, xdebug_sprintf("<small>boolean</small> <font color='%s'>%s</font>", COLOR_BOOL, Z_TYPE_P(*struc) == IS_TRUE ? "true" : "false"), 1);
+			xdebug_str_add_fmt(str, "<small>boolean</small> <font color='%s'>false</font>", COLOR_BOOL);
 			break;
 
 		case IS_NULL:
-			xdebug_str_add(str, xdebug_sprintf("<font color='%s'>null</font>", COLOR_NULL), 1);
+			xdebug_str_add_fmt(str, "<font color='%s'>null</font>", COLOR_NULL);
 			break;
 
 		case IS_LONG:
-			xdebug_str_add(str, xdebug_sprintf("<small>int</small> <font color='%s'>" XDEBUG_INT_FMT "</font>", COLOR_LONG, Z_LVAL_P(*struc)), 1);
+			xdebug_str_add_fmt(str, "<small>int</small> <font color='%s'>" XDEBUG_INT_FMT "</font>", COLOR_LONG, Z_LVAL_P(*struc));
 			break;
 
 		case IS_DOUBLE:
-			xdebug_str_add(str, xdebug_sprintf("<small>float</small> <font color='%s'>%.*G</font>", COLOR_DOUBLE, (int) EG(precision), Z_DVAL_P(*struc)), 1);
+			xdebug_str_add_fmt(str, "<small>float</small> <font color='%s'>%.*G</font>", COLOR_DOUBLE, (int) EG(precision), Z_DVAL_P(*struc));
 			break;
 
 		case IS_STRING:
-			xdebug_str_add(str, xdebug_sprintf("<small>string</small> <font color='%s'>'", COLOR_STRING), 1);
+			xdebug_str_add_fmt(str, "<small>string</small> <font color='%s'>'", COLOR_STRING);
 			if ((size_t) Z_STRLEN_P(*struc) > (size_t) options->max_data) {
 				tmp_str = xdebug_xmlize(Z_STRVAL_P(*struc), options->max_data, &newlen);
 				xdebug_str_addl(str, tmp_str, newlen, 0);
 				efree(tmp_str);
-				xdebug_str_addl(str, "'...</font>", 11, 0);
+				xdebug_str_add_literal(str, "'...</font>");
 			} else {
 				tmp_str = xdebug_xmlize(Z_STRVAL_P(*struc), Z_STRLEN_P(*struc), &newlen);
 				xdebug_str_addl(str, tmp_str, newlen, 0);
 				efree(tmp_str);
-				xdebug_str_addl(str, "'</font>", 8, 0);
+				xdebug_str_add_literal(str, "'</font>");
 			}
-			xdebug_str_add(str, xdebug_sprintf(" <i>(length=%d)</i>", Z_STRLEN_P(*struc)), 1);
+			xdebug_str_add_fmt(str, " <i>(length=%d)</i>", Z_STRLEN_P(*struc));
 			break;
 
 		case IS_ARRAY:
 			myht = Z_ARRVAL_P(*struc);
-			xdebug_str_add(str, xdebug_sprintf("\n%*s", (level - 1) * 4, ""), 1);
+			xdebug_str_add_fmt(str, "\n%*s", (level - 1) * 4, "");
 
 			if (!xdebug_zend_hash_is_recursive(myht)) {
-				xdebug_str_add(str, xdebug_sprintf("<b>array</b> <i>(size=%d)</i>\n", myht->nNumOfElements), 1);
+				xdebug_str_add_fmt(str, "<b>array</b> <i>(size=%d)</i>\n", myht->nNumOfElements);
 				if (level <= options->max_depth) {
 					if (myht->nNumOfElements) {
 						options->runtime[level].current_element_nr = 0;
@@ -194,14 +203,14 @@ void xdebug_var_export_html(zval **struc, xdebug_str *str, int level, int debug_
 
 						xdebug_zend_hash_apply_protection_end(myht);
 					} else {
-						xdebug_str_add(str, xdebug_sprintf("%*s", (level * 4) - 2, ""), 1);
-						xdebug_str_add(str, xdebug_sprintf("<i><font color='%s'>empty</font></i>\n", COLOR_EMPTY), 1);
+						xdebug_str_add_fmt(str, "%*s", (level * 4) - 2, "");
+						xdebug_str_add_fmt(str, "<i><font color='%s'>empty</font></i>\n", COLOR_EMPTY);
 					}
 				} else {
-					xdebug_str_add(str, xdebug_sprintf("%*s...\n", (level * 4) - 2, ""), 1);
+					xdebug_str_add_fmt(str, "%*s...\n", (level * 4) - 2, "");
 				}
 			} else {
-				xdebug_str_addl(str, "<i>&amp;</i><b>array</b>\n", 21, 0);
+				xdebug_str_add_literal(str, "<i>&amp;</i><b>array</b>\n");
 			}
 			break;
 
@@ -211,14 +220,15 @@ void xdebug_var_export_html(zval **struc, xdebug_str *str, int level, int debug_
 #else
 			myht = xdebug_objdebug_pp(struc, &is_temp);
 #endif
-			xdebug_str_add(str, xdebug_sprintf("\n%*s", (level - 1) * 4, ""), 1);
+			xdebug_str_add_fmt(str, "\n%*s", (level - 1) * 4, "");
 
-			if (!xdebug_zend_hash_is_recursive(myht)) {
-				char *class_name = (char*) STR_NAME_VAL(Z_OBJCE_P(*struc)->name);
-				xdebug_str_add(str, xdebug_sprintf("<b>object</b>(<i>%s</i>)", class_name), 1);
-				xdebug_str_add(str, xdebug_sprintf("[<i>%d</i>]\n", Z_OBJ_HANDLE_P(*struc)), 1);
+			if (!myht || !xdebug_zend_hash_is_recursive(myht)) {
+				xdebug_str_add_literal(str, "<b>object</b>(<i>");
+				xdebug_str_add(str, ZSTR_VAL(Z_OBJCE_P(*struc)->name), 0);
+				xdebug_str_add_literal(str, "</i>)");
+				xdebug_str_add_fmt(str, "[<i>%d</i>]\n", Z_OBJ_HANDLE_P(*struc));
 
-				if (level <= options->max_depth) {
+				if (myht && (level <= options->max_depth)) {
 					options->runtime[level].current_element_nr = 0;
 					options->runtime[level].start_element_nr = 0;
 					options->runtime[level].end_element_nr = options->max_children;
@@ -226,16 +236,18 @@ void xdebug_var_export_html(zval **struc, xdebug_str *str, int level, int debug_
 					xdebug_zend_hash_apply_protection_begin(myht);
 
 					ZEND_HASH_FOREACH_KEY_VAL(myht, num, key, val) {
-						xdebug_object_element_export_html(*struc, val, num, key, level, str, debug_zval, options, class_name);
+						xdebug_object_element_export_html(*struc, val, num, key, level, str, debug_zval, options, ZSTR_VAL(Z_OBJCE_P(*struc)->name));
 					} ZEND_HASH_FOREACH_END();
 
 					xdebug_zend_hash_apply_protection_end(myht);
 				} else {
-					xdebug_str_add(str, xdebug_sprintf("%*s...\n", (level * 4) - 2, ""), 1);
+					xdebug_str_add_fmt(str, "%*s...\n", (level * 4) - 2, "");
 				}
 			} else {
-				xdebug_str_add(str, xdebug_sprintf("<i>&amp;</i><b>object</b>(<i>%s</i>)", STR_NAME_VAL(Z_OBJCE_P(*struc)->name)), 1);
-				xdebug_str_add(str, xdebug_sprintf("[<i>%d</i>]\n", Z_OBJ_HANDLE_P(*struc)), 1);
+				xdebug_str_add_literal(str, "<i>&amp;</i><b>object</b>(<i>");
+				xdebug_str_add(str, ZSTR_VAL(Z_OBJCE_P(*struc)->name), 0);
+				xdebug_str_add_literal(str, "</i>)");
+				xdebug_str_add_fmt(str, "[<i>%d</i>]\n", Z_OBJ_HANDLE_P(*struc));
 			}
 #if PHP_VERSION_ID >= 70400
 			zend_release_properties(myht);
@@ -248,20 +260,20 @@ void xdebug_var_export_html(zval **struc, xdebug_str *str, int level, int debug_
 			char *type_name;
 
 			type_name = (char *) zend_rsrc_list_get_rsrc_type(Z_RES_P(*struc));
-			xdebug_str_add(str, xdebug_sprintf("<b>resource</b>(<i>%ld</i><font color='%s'>,</font> <i>%s</i>)", Z_RES_P(*struc)->handle, COLOR_RESOURCE, type_name ? type_name : "Unknown"), 1);
+			xdebug_str_add_fmt(str, "<b>resource</b>(<i>%ld</i><font color='%s'>,</font> <i>%s</i>)", Z_RES_P(*struc)->handle, COLOR_RESOURCE, type_name ? type_name : "Unknown");
 			break;
 		}
 
 		case IS_UNDEF:
-			xdebug_str_add(str, xdebug_sprintf("<font color='%s'>*uninitialized*</font>", COLOR_NULL), 0);
+			xdebug_str_add_fmt(str, "<font color='%s'>*uninitialized*</font>", COLOR_NULL);
 			break;
 
 		default:
-			xdebug_str_add(str, xdebug_sprintf("<font color='%s'>NFC</font>", COLOR_NULL), 0);
+			xdebug_str_add_fmt(str, "<font color='%s'>NFC</font>", COLOR_NULL);
 			break;
 	}
-	if (Z_TYPE_P(*struc) != IS_ARRAY && Z_TYPE_P(*struc) != IS_OBJECT) {
-		xdebug_str_addl(str, "\n", 1, 0);
+	if (z_type != IS_ARRAY && z_type != IS_OBJECT) {
+		xdebug_str_addc(str, '\n');
 	}
 }
 
@@ -275,25 +287,25 @@ xdebug_str* xdebug_get_zval_value_html(char *name, zval *val, int debug_zval, xd
 		default_options = 1;
 	}
 
-	xdebug_str_addl(str, "<pre class='xdebug-var-dump' dir='ltr'>", 39, 0);
+	xdebug_str_add_literal(str, "<pre class='xdebug-var-dump' dir='ltr'>");
 	if (options->show_location && !debug_zval) {
 		char *formatted_filename;
-		xdebug_format_filename(&formatted_filename, XINI_BASE(filename_format), "%f", zend_get_executed_filename());
+		xdebug_format_filename(&formatted_filename, "%f", zend_get_executed_filename_ex());
 
-		if (strlen(XINI_BASE(file_link_format)) > 0) {
+		if (strlen(XINI_LIB(file_link_format)) > 0) {
 			char *file_link;
 
 			xdebug_format_file_link(&file_link, zend_get_executed_filename(), zend_get_executed_lineno());
-			xdebug_str_add(str, xdebug_sprintf("\n<small><a href='%s'>%s:%d</a>:</small>", file_link, formatted_filename, zend_get_executed_lineno()), 1);
+			xdebug_str_add_fmt(str, "\n<small><a href='%s'>%s:%d</a>:</small>", file_link, formatted_filename, zend_get_executed_lineno());
 			xdfree(file_link);
 		} else {
-			xdebug_str_add(str, xdebug_sprintf("\n<small>%s:%d:</small>", formatted_filename, zend_get_executed_lineno()), 1);
+			xdebug_str_add_fmt(str, "\n<small>%s:%d:</small>", formatted_filename, zend_get_executed_lineno());
 		}
 
 		xdfree(formatted_filename);
 	}
 	xdebug_var_export_html(&val, str, 1, debug_zval, options);
-	xdebug_str_addl(str, "</pre>", 6, 0);
+	xdebug_str_add_literal(str, "</pre>");
 
 	if (default_options) {
 		xdfree(options->runtime);
@@ -307,62 +319,66 @@ static void xdebug_var_synopsis_html(zval **struc, xdebug_str *str, int level, i
 {
 	HashTable *myht;
 	zval *tmpz;
+	int   z_type;
+
+	z_type = Z_TYPE_P(*struc);
 
 	if (debug_zval) {
 		xdebug_add_variable_attributes(str, *struc, XDEBUG_VAR_ATTR_HTML);
 	}
-	if (Z_TYPE_P(*struc) == IS_REFERENCE) {
+	if (z_type == IS_REFERENCE) {
 		tmpz = &((*struc)->value.ref->val);
 		struc = &tmpz;
 	}
 
-	switch (Z_TYPE_P(*struc)) {
+	switch (z_type) {
 		case IS_TRUE:
+			xdebug_str_add_fmt(str, "<font color='%s'>true</font>", COLOR_BOOL);
+			break;
+
 		case IS_FALSE:
-			xdebug_str_add(str, xdebug_sprintf("<font color='%s'>%s</font>", COLOR_BOOL, Z_TYPE_P(*struc) == IS_TRUE ? "true" : "false"), 1);
+			xdebug_str_add_fmt(str, "<font color='%s'>false</font>", COLOR_BOOL);
 			break;
 
 		case IS_NULL:
-			xdebug_str_add(str, xdebug_sprintf("<font color='%s'>null</font>", COLOR_NULL), 1);
+			xdebug_str_add_fmt(str, "<font color='%s'>null</font>", COLOR_NULL);
 			break;
 
 		case IS_LONG:
-			xdebug_str_add(str, xdebug_sprintf("<font color='%s'>long</font>", COLOR_LONG), 1);
+			xdebug_str_add_fmt(str, "<font color='%s'>long</font>", COLOR_LONG);
 			break;
 
 		case IS_DOUBLE:
-			xdebug_str_add(str, xdebug_sprintf("<font color='%s'>double</font>", COLOR_DOUBLE), 1);
+			xdebug_str_add_fmt(str, "<font color='%s'>double</font>", COLOR_DOUBLE);
 			break;
 
 		case IS_STRING:
-			xdebug_str_add(str, xdebug_sprintf("<font color='%s'>string(%d)</font>", COLOR_STRING, Z_STRLEN_P(*struc)), 1);
+			xdebug_str_add_fmt(str, "<font color='%s'>string(%d)</font>", COLOR_STRING, Z_STRLEN_P(*struc));
 			break;
 
 		case IS_ARRAY:
 			myht = Z_ARRVAL_P(*struc);
-			xdebug_str_add(str, xdebug_sprintf("<font color='%s'>array(%d)</font>", COLOR_ARRAY, myht->nNumOfElements), 1);
+			xdebug_str_add_fmt(str, "<font color='%s'>array(%d)</font>", COLOR_ARRAY, myht->nNumOfElements);
 			break;
 
 		case IS_OBJECT:
-			xdebug_str_add(str, xdebug_sprintf("<font color='%s'>object(%s)", COLOR_OBJECT, STR_NAME_VAL(Z_OBJCE_P(*struc)->name)), 1);
-			xdebug_str_add(str, xdebug_sprintf("[%d]", Z_OBJ_HANDLE_P(*struc)), 1);
-			xdebug_str_addl(str, "</font>", 7, 0);
+			xdebug_str_add_fmt(str, "<font color='%s'>object(%s)[%d]</font>", COLOR_OBJECT, ZSTR_VAL(Z_OBJCE_P(*struc)->name), Z_OBJ_HANDLE_P(*struc));
 			break;
 
 		case IS_RESOURCE: {
 			char *type_name;
 
 			type_name = (char *) zend_rsrc_list_get_rsrc_type(Z_RES_P(*struc));
-			xdebug_str_add(str, xdebug_sprintf("<font color='%s'>resource(%ld, %s)</font>", COLOR_RESOURCE, Z_RES_P(*struc)->handle, type_name ? type_name : "Unknown"), 1);
+			xdebug_str_add_fmt(str, "<font color='%s'>resource(%ld, %s)</font>", COLOR_RESOURCE, Z_RES_P(*struc)->handle, type_name ? type_name : "Unknown");
 			break;
 		}
 
 		case IS_UNDEF:
-			xdebug_str_add(str, xdebug_sprintf("<font color='%s'>*uninitialized*</font>", COLOR_NULL), 0);
+			xdebug_str_add_fmt(str, "<font color='%s'>*uninitialized*</font>", COLOR_NULL);
 			break;
 
 		default:
-			xdebug_str_add(str, xdebug_sprintf("<font color='%s'>NFC</font>", COLOR_NULL), 0);
+			xdebug_str_add_fmt(str, "<font color='%s'>NFC</font>", COLOR_NULL);
 			break;
 	}
 }
