@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | Xdebug                                                               |
    +----------------------------------------------------------------------+
-   | Copyright (c) 2002-2019 Derick Rethans                               |
+   | Copyright (c) 2002-2020 Derick Rethans                               |
    +----------------------------------------------------------------------+
    | This source file is subject to version 1.01 of the Xdebug license,   |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -25,7 +25,7 @@
 #include "trace_html.h"
 
 #include "lib/compat.h"
-#include "lib/private.h"
+#include "lib/log.h"
 #include "lib/str.h"
 #include "lib/var_export_line.h"
 
@@ -54,38 +54,52 @@ static xdebug_trace_handler_t *xdebug_select_trace_handler(int options)
 	return tmp;
 }
 
-FILE *xdebug_trace_open_file(char *fname, char *script_filename, long options, char **used_fname)
+FILE *xdebug_trace_open_file(char *requested_filename, zend_string *script_filename, long options, char **used_fname)
 {
 	FILE *file;
-	char *filename;
+	char *filename_to_use;
+	char *generated_filename = NULL;
+	char *output_dir = xdebug_lib_get_output_dir(); /* not duplicated */
 
-	if (fname && strlen(fname)) {
-		filename = xdstrdup(fname);
+	if (requested_filename && strlen(requested_filename)) {
+		filename_to_use = xdstrdup(requested_filename);
 	} else {
+
 		if (!strlen(XINI_TRACE(trace_output_name)) ||
-			xdebug_format_output_filename(&fname, XINI_TRACE(trace_output_name), script_filename) <= 0
+			xdebug_format_output_filename(&generated_filename, XINI_TRACE(trace_output_name), ZSTR_VAL(script_filename)) <= 0
 		) {
 			/* Invalid or empty xdebug.trace_output_name */
 			return NULL;
 		}
-		if (IS_SLASH(XINI_TRACE(trace_output_dir)[strlen(XINI_TRACE(trace_output_dir)) - 1])) {
-			filename = xdebug_sprintf("%s%s", XINI_TRACE(trace_output_dir), fname);
+
+		/* Add a slash if none is present in the output_dir setting */
+		output_dir = xdebug_lib_get_output_dir(); /* not duplicated */
+
+		if (IS_SLASH(output_dir[strlen(output_dir) - 1])) {
+			filename_to_use = xdebug_sprintf("%s%s", output_dir, generated_filename);
 		} else {
-			filename = xdebug_sprintf("%s%c%s", XINI_TRACE(trace_output_dir), DEFAULT_SLASH, fname);
+			filename_to_use = xdebug_sprintf("%s%c%s", output_dir, DEFAULT_SLASH, generated_filename);
 		}
-		xdfree(fname);
 	}
 	if (options & XDEBUG_TRACE_OPTION_APPEND) {
-		file = xdebug_fopen(filename, "a", (options & XDEBUG_TRACE_OPTION_NAKED_FILENAME) ? NULL : "xt", used_fname);
+		file = xdebug_fopen(filename_to_use, "a", (options & XDEBUG_TRACE_OPTION_NAKED_FILENAME) ? NULL : "xt", used_fname);
 	} else {
-		file = xdebug_fopen(filename, "w", (options & XDEBUG_TRACE_OPTION_NAKED_FILENAME) ? NULL : "xt", used_fname);
+		file = xdebug_fopen(filename_to_use, "w", (options & XDEBUG_TRACE_OPTION_NAKED_FILENAME) ? NULL : "xt", used_fname);
 	}
-	xdfree(filename);
+
+	if (!file) {
+		xdebug_log_diagnose_permissions(XLOG_CHAN_TRACE, output_dir, generated_filename);
+	}
+
+	if (generated_filename) {
+		xdfree(generated_filename);
+	}
+	xdfree(filename_to_use);
 
 	return file;
 }
 
-static char* xdebug_start_trace(char* fname, char *script_filename, long options)
+static char* xdebug_start_trace(char* fname, zend_string *script_filename, long options)
 {
 	if (XG_TRACE(trace_context)) {
 		return NULL;
@@ -94,21 +108,32 @@ static char* xdebug_start_trace(char* fname, char *script_filename, long options
 	XG_TRACE(trace_handler) = xdebug_select_trace_handler(options);
 	XG_TRACE(trace_context) = (void*) XG_TRACE(trace_handler)->init(fname, script_filename, options);
 
-	if (XG_TRACE(trace_context)) {
-		XG_TRACE(trace_handler)->write_header(XG_TRACE(trace_context));
-		return xdstrdup(XG_TRACE(trace_handler)->get_filename(XG_TRACE(trace_context)));
+	if (!XG_TRACE(trace_context)) {
+		return NULL;
 	}
 
-	return NULL;
+	XG_TRACE(trace_handler)->write_header(XG_TRACE(trace_context));
+	return xdstrdup(XG_TRACE(trace_handler)->get_filename(XG_TRACE(trace_context)));
 }
 
 static void xdebug_stop_trace(void)
 {
-	if (XG_TRACE(trace_context)) {
-		XG_TRACE(trace_handler)->write_footer(XG_TRACE(trace_context));
-		XG_TRACE(trace_handler)->deinit(XG_TRACE(trace_context));
-		XG_TRACE(trace_context) = NULL;
+	if (!XG_TRACE(trace_context)) {
+		return;
 	}
+
+	XG_TRACE(trace_handler)->write_footer(XG_TRACE(trace_context));
+	XG_TRACE(trace_handler)->deinit(XG_TRACE(trace_context));
+	XG_TRACE(trace_context) = NULL;
+}
+
+char *xdebug_get_trace_filename(void)
+{
+	if (!(XG_TRACE(trace_context) && XG_TRACE(trace_handler) && XG_TRACE(trace_handler)->get_filename)) {
+		return NULL;
+	}
+
+	return XG_TRACE(trace_handler)->get_filename(XG_TRACE(trace_context));
 }
 
 PHP_FUNCTION(xdebug_start_trace)
@@ -117,87 +142,57 @@ PHP_FUNCTION(xdebug_start_trace)
 	size_t fname_len = 0;
 	char *trace_fname;
 	zend_long options = XINI_TRACE(trace_options);
+	function_stack_entry *fse;
 
-	if (!XG_TRACE(trace_context)) {
-		function_stack_entry *fse;
+	WARN_AND_RETURN_IF_MODE_IS_NOT(XDEBUG_MODE_TRACING);
 
-		if (zend_parse_parameters(ZEND_NUM_ARGS(), "|sl", &fname, &fname_len, &options) == FAILURE) {
-			return;
-		}
-
-		fse = xdebug_get_stack_frame(0);
-
-		if ((trace_fname = xdebug_start_trace(fname, fse->filename, options)) != NULL) {
-			RETVAL_STRING(trace_fname);
-			xdfree(trace_fname);
-			return;
-		} else {
-			php_error(E_NOTICE, "Trace could not be started");
-		}
-
-		RETURN_FALSE;
-	} else {
+	if (XG_TRACE(trace_context)) {
 		php_error(E_NOTICE, "Function trace already started");
 		RETURN_FALSE;
 	}
+
+	if (zend_parse_parameters(ZEND_NUM_ARGS(), "|sl", &fname, &fname_len, &options) == FAILURE) {
+		return;
+	}
+
+	fse = xdebug_get_stack_frame(0);
+
+	if ((trace_fname = xdebug_start_trace(fname, fse->filename, options)) != NULL) {
+		RETVAL_STRING(trace_fname);
+		xdfree(trace_fname);
+		return;
+	} else {
+		php_error(E_NOTICE, "Trace could not be started");
+	}
+
+	RETURN_FALSE;
 }
 
 PHP_FUNCTION(xdebug_stop_trace)
 {
-	if (XG_TRACE(trace_context)) {
-		RETVAL_STRING(XG_TRACE(trace_handler)->get_filename(XG_TRACE(trace_context)));
-		xdebug_stop_trace();
-	} else {
+	WARN_AND_RETURN_IF_MODE_IS_NOT(XDEBUG_MODE_TRACING);
+
+	if (!XG_TRACE(trace_context)) {
 		RETVAL_FALSE;
 		php_error(E_NOTICE, "Function trace was not started");
 	}
+
+	RETVAL_STRING(XG_TRACE(trace_handler)->get_filename(XG_TRACE(trace_context)));
+	xdebug_stop_trace();
 }
 
 PHP_FUNCTION(xdebug_get_tracefile_name)
 {
-	if (XG_TRACE(trace_context) && XG_TRACE(trace_handler) && XG_TRACE(trace_handler)->get_filename) {
-		RETVAL_STRING(XG_TRACE(trace_handler)->get_filename(XG_TRACE(trace_context)));
-	} else {
+	char *filename;
+
+	WARN_AND_RETURN_IF_MODE_IS_NOT(XDEBUG_MODE_TRACING);
+
+	filename = xdebug_get_trace_filename();
+	if (!filename) {
 		RETURN_FALSE;
 	}
-}
 
-static int xdebug_include_or_eval_handler(zend_execute_data *execute_data)
-{
-	zend_op_array *op_array = &execute_data->func->op_array;
-	const zend_op *opline = execute_data->opline;
-
-	xdebug_coverage_record_include_if_active(execute_data, op_array);
-	if (opline->extended_value == ZEND_EVAL) {
-		zval *inc_filename;
-		zval tmp_inc_filename;
-		int  is_var;
-
-		inc_filename = xdebug_get_zval(execute_data, opline->op1_type, &opline->op1, &is_var);
-
-		/* If there is no inc_filename, we're just bailing out instead */
-		if (!inc_filename) {
-			return ZEND_USER_OPCODE_DISPATCH;
-		}
-
-		if (Z_TYPE_P(inc_filename) != IS_STRING) {
-			tmp_inc_filename = *inc_filename;
-			zval_copy_ctor(&tmp_inc_filename);
-			convert_to_string(&tmp_inc_filename);
-			inc_filename = &tmp_inc_filename;
-		}
-
-		/* Now let's store this info */
-		if (XG_BASE(last_eval_statement)) {
-			efree(XG_BASE(last_eval_statement));
-		}
-		XG_BASE(last_eval_statement) = estrndup(Z_STRVAL_P(inc_filename), Z_STRLEN_P(inc_filename));
-
-		if (inc_filename == &tmp_inc_filename) {
-			zval_dtor(&tmp_inc_filename);
-		}
-	}
-	return ZEND_USER_OPCODE_DISPATCH;
+	RETVAL_STRING(filename);
 }
 
 #if PHP_VERSION_ID >= 70400
@@ -238,7 +233,6 @@ static int xdebug_is_static_call(const zend_op *cur_opcode, const zend_op *prev_
 		return 1;
 	}
 # endif
-# if PHP_VERSION_ID >= 70100
 	while (!(opcode_ptr->opcode == ZEND_EXT_STMT) && !((opcode_ptr->opcode == ZEND_FETCH_STATIC_PROP_W) || (opcode_ptr->opcode == ZEND_FETCH_STATIC_PROP_RW))) {
 		opcode_ptr = opcode_ptr - 1;
 	}
@@ -246,37 +240,33 @@ static int xdebug_is_static_call(const zend_op *cur_opcode, const zend_op *prev_
 		*found_opcode = opcode_ptr;
 		return 1;
 	}
-# else
-	while (!(opcode_ptr->opcode == ZEND_EXT_STMT) && !((opcode_ptr->opcode == ZEND_FETCH_W) || (opcode_ptr->opcode == ZEND_FETCH_RW))) {
-		opcode_ptr = opcode_ptr - 1;
-	}
-	if (((opcode_ptr->opcode == ZEND_FETCH_W) || (opcode_ptr->opcode == ZEND_FETCH_RW)) && opcode_ptr->extended_value == ZEND_FETCH_STATIC_MEMBER) {
-		*found_opcode = opcode_ptr;
-		return 1;
-	}
-# endif
 	return 0;
 }
 
 static const zend_op *xdebug_find_referenced_opline(zend_execute_data *execute_data, const zend_op *cur_opcode, int op1_or_op2)
 {
+	size_t variable_number;
+	const zend_op *scan_opcode;
+	int found;
 	int op_type = (op1_or_op2 == 1) ? cur_opcode->op1_type : cur_opcode->op2_type;
 
-	if (op_type == IS_VAR) {
-		size_t variable_number = (op1_or_op2 == 1) ? cur_opcode->op1.var : cur_opcode->op2.var;
-		const zend_op *scan_opcode = cur_opcode;
-		int found = 0;
-
-		/* Scroll up until we find a RES of IS_VAR with the right value */
-		do {
-			scan_opcode--;
-			if (scan_opcode->result_type == IS_VAR && scan_opcode->result.var == variable_number) {
-				found = 1;
-			}
-		} while (!found);
-		return scan_opcode;
+	if (op_type != IS_VAR) {
+		return NULL;
 	}
-	return NULL;
+
+	variable_number = (op1_or_op2 == 1) ? cur_opcode->op1.var : cur_opcode->op2.var;
+	scan_opcode = cur_opcode;
+	found = 0;
+
+	/* Scroll up until we find a RES of IS_VAR with the right value */
+	do {
+		scan_opcode--;
+		if (scan_opcode->result_type == IS_VAR && scan_opcode->result.var == variable_number) {
+			found = 1;
+		}
+	} while (!found);
+
+	return scan_opcode;
 }
 
 static int is_fetch_op(const zend_op *op)
@@ -304,11 +294,8 @@ static char *xdebug_find_var_name(zend_execute_data *execute_data, const zend_op
 	prev_opcode = cur_opcode - 1;
 
 	if (cur_opcode->opcode == ZEND_QM_ASSIGN) {
-#if PHP_VERSION_ID >= 70000
-		xdebug_str_add(&name, xdebug_sprintf("$%s", zend_get_compiled_variable_name(op_array, cur_opcode->result.var)->val), 1);
-#else
-		xdebug_str_add(&name, xdebug_sprintf("$%s", zend_get_compiled_variable_name(op_array, cur_opcode->result.var, &cv_len)), 1);
-#endif
+		xdebug_str_addc(&name, '$');
+		xdebug_str_add(&name, zend_get_compiled_variable_name(op_array, cur_opcode->result.var)->val, 0);
 	}
 
 	is_static = xdebug_is_static_call(cur_opcode, prev_opcode, &static_opcode_ptr);
@@ -317,11 +304,12 @@ static char *xdebug_find_var_name(zend_execute_data *execute_data, const zend_op
 
 	if (cur_opcode->op1_type == IS_CV) {
 		if (!lower_bound) {
-			xdebug_str_add(&name, xdebug_sprintf("$%s", zend_get_compiled_variable_name(op_array, cur_opcode->op1.var)->val), 1);
+			xdebug_str_addc(&name, '$');
+			xdebug_str_add(&name, zend_get_compiled_variable_name(op_array, cur_opcode->op1.var)->val, 0);
 		}
 	} else if (cur_opcode->op1_type == IS_VAR && cur_opcode->opcode == ZEND_ASSIGN && (prev_opcode->opcode == ZEND_FETCH_W || prev_opcode->opcode == ZEND_FETCH_RW)) {
 		if (is_static) {
-			xdebug_str_add(&name, xdebug_sprintf("self::"), 1);
+			xdebug_str_add_literal(&name, "self::");
 		} else {
 			zval_value = xdebug_get_zval_value_line(xdebug_get_zval_with_opline(execute_data, prev_opcode, prev_opcode->op1_type, &prev_opcode->op1, &is_var), 0, options);
 			xdebug_str_addc(&name, '$');
@@ -329,11 +317,11 @@ static char *xdebug_find_var_name(zend_execute_data *execute_data, const zend_op
 			xdebug_str_free(zval_value);
 		}
 	} else if (is_static) { /* todo : see if you can change this and the previous cases around */
-		xdebug_str_add(&name, xdebug_sprintf("self::"), 1 );
+		xdebug_str_add_literal(&name, "self::");
 	}
 	if (cur_opcode->opcode >= ZEND_PRE_INC_OBJ && cur_opcode->opcode <= ZEND_POST_DEC_OBJ) {
 		zval_value = xdebug_get_zval_value_line(xdebug_get_zval(execute_data, cur_opcode->op2_type, &cur_opcode->op2, &is_var), 0, options);
-		xdebug_str_addl(&name, "$this->", 7, 0);
+		xdebug_str_add_literal(&name, "$this->");
 		xdebug_str_add_str(&name, zval_value);
 		xdebug_str_free(zval_value);
 	}
@@ -372,18 +360,17 @@ static char *xdebug_find_var_name(zend_execute_data *execute_data, const zend_op
 				opcode_ptr->op1_type == IS_UNUSED &&
 				(opcode_ptr->opcode == ZEND_FETCH_OBJ_W || opcode_ptr->opcode == ZEND_FETCH_OBJ_RW)
 			) {
-				xdebug_str_add(&name, "$this", 0);
+				xdebug_str_add_literal(&name, "$this");
 			}
 			if (opcode_ptr->op1_type == IS_CV) {
-				xdebug_str_add(&name, xdebug_sprintf("$%s", zend_get_compiled_variable_name(op_array, opcode_ptr->op1.var)->val), 1);
+				xdebug_str_addc(&name, '$');
+				xdebug_str_add(&name, zend_get_compiled_variable_name(op_array, opcode_ptr->op1.var)->val, 0);
 			}
-#if PHP_VERSION_ID >= 70100
 			if (opcode_ptr->opcode == ZEND_FETCH_STATIC_PROP_W || opcode_ptr->opcode == ZEND_FETCH_STATIC_PROP_R || opcode_ptr->opcode == ZEND_FETCH_STATIC_PROP_RW) {
 				zval_value = xdebug_get_zval_value_line(xdebug_get_zval_with_opline(execute_data, opcode_ptr, opcode_ptr->op1_type, &opcode_ptr->op1, &is_var), 0, options);
 				xdebug_str_add_str(&name, zval_value);
 				xdebug_str_free(zval_value);
 			}
-#endif
 			if (opcode_ptr->opcode == ZEND_FETCH_W) {
 				zval_value = xdebug_get_zval_value_line(xdebug_get_zval_with_opline(execute_data, opcode_ptr, opcode_ptr->op1_type, &opcode_ptr->op1, &is_var), 0, options);
 				xdebug_str_add_str(&name, zval_value);
@@ -404,20 +391,11 @@ static char *xdebug_find_var_name(zend_execute_data *execute_data, const zend_op
 				xdebug_str_free(zval_value);
 			} else if (opcode_ptr->opcode == ZEND_FETCH_OBJ_W || opcode_ptr->opcode == ZEND_FETCH_OBJ_RW) {
 				zval_value = xdebug_get_zval_value_line(xdebug_get_zval_with_opline(execute_data, opcode_ptr, opcode_ptr->op2_type, &opcode_ptr->op2, &is_var), 0, options);
-				xdebug_str_addl(&name, "->", 2, 0);
+				xdebug_str_add_literal(&name, "->");
 				xdebug_str_add_str(&name, zval_value);
 				xdebug_str_free(zval_value);
 			}
 			opcode_ptr = opcode_ptr + 1;
-#if PHP_VERSION_ID < 70100
-			/* This is a hack, as we really ought to continue until we find the
-			 * right temp variable number in the "return_value" part of the
-			 * opcode. But this is only needed for PHP 7.0 so we just cop out
-			 * after the first one */
-			if (is_static && lower_bound && lower_bound->opcode == ZEND_ASSIGN_REF) {
-				cv_found = 1;
-			}
-#endif
 			if (opcode_ptr->op1_type == IS_CV) {
 				cv_found = 1;
 			}
@@ -431,15 +409,16 @@ static char *xdebug_find_var_name(zend_execute_data *execute_data, const zend_op
 #endif
 	) {
 		if (cur_opcode->op1_type == IS_UNUSED) {
-			xdebug_str_add(&name, "$this", 0);
+			xdebug_str_add_literal(&name, "$this");
 		}
 		dimval = xdebug_get_zval(execute_data, cur_opcode->op2_type, &cur_opcode->op2, &is_var);
-		xdebug_str_add(&name, xdebug_sprintf("->%s", Z_STRVAL_P(dimval)), 1);
+		xdebug_str_add_literal(&name, "->");
+		xdebug_str_add(&name, Z_STRVAL_P(dimval), 0);
 	}
 #if PHP_VERSION_ID >= 70400
 	if (cur_opcode->opcode == ZEND_ASSIGN_STATIC_PROP_REF) {
 		dimval = xdebug_get_zval(execute_data, cur_opcode->op1_type, &cur_opcode->op1, &is_var);
-		xdebug_str_add(&name, xdebug_sprintf("%s", Z_STRVAL_P(dimval)), 1);
+		xdebug_str_add(&name, Z_STRVAL_P(dimval), 0);
 	}
 	if (cur_opcode->opcode == ZEND_ASSIGN_DIM_OP) {
 		zval_value = xdebug_get_zval_value_line(xdebug_get_zval(execute_data, cur_opcode->op2_type, &cur_opcode->op2, &is_var), 0, NULL);
@@ -451,16 +430,16 @@ static char *xdebug_find_var_name(zend_execute_data *execute_data, const zend_op
 	if (cur_opcode->opcode == ZEND_ASSIGN_OBJ_OP) {
 		zval_value = xdebug_get_zval_value_line(xdebug_get_zval(execute_data, cur_opcode->op2_type, &cur_opcode->op2, &is_var), 0, options);
 		if (cur_opcode->op1_type == IS_UNUSED) {
-			xdebug_str_addl(&name, "$this->", 7, 0);
+			xdebug_str_add_literal(&name, "$this->");
 		} else {
-			xdebug_str_addl(&name, "->", 2, 0);
+			xdebug_str_add_literal(&name, "->");
 		}
 		xdebug_str_add_str(&name, zval_value);
 		xdebug_str_free(zval_value);
 	}
 	if (cur_opcode->opcode == ZEND_ASSIGN_STATIC_PROP_OP) {
 		zval_value = xdebug_get_zval_value_line(xdebug_get_zval(execute_data, cur_opcode->op1_type, &cur_opcode->op1, &is_var), 0, options);
-		xdebug_str_addl(&name, "self::", 6, 0);
+		xdebug_str_add_literal(&name, "self::");
 		xdebug_str_add_str(&name, zval_value);
 		xdebug_str_free(zval_value);
 	}
@@ -478,9 +457,9 @@ static char *xdebug_find_var_name(zend_execute_data *execute_data, const zend_op
 		if (cur_opcode->extended_value == ZEND_ASSIGN_OBJ) {
 			zval_value = xdebug_get_zval_value_line(xdebug_get_zval(execute_data, cur_opcode->op2_type, &cur_opcode->op2, &is_var), 0, options);
 			if (cur_opcode->op1_type == IS_UNUSED) {
-				xdebug_str_addl(&name, "$this->", 7, 0);
+				xdebug_str_add_literal(&name, "$this->");
 			} else {
-				xdebug_str_addl(&name, "->", 2, 0);
+				xdebug_str_add_literal(&name, "->");
 			}
 			xdebug_str_add_str(&name, zval_value);
 			xdebug_str_free(zval_value);
@@ -490,7 +469,7 @@ static char *xdebug_find_var_name(zend_execute_data *execute_data, const zend_op
 
 	if (cur_opcode->opcode == ZEND_ASSIGN_DIM) {
 		if (next_opcode->opcode == ZEND_OP_DATA && cur_opcode->op2_type == IS_UNUSED) {
-			xdebug_str_add(&name, "[]", 0);
+			xdebug_str_add_literal(&name, "[]");
 		} else {
 			zval_value = xdebug_get_zval_value_line(xdebug_get_zval_with_opline(execute_data, opcode_ptr, opcode_ptr->op2_type, &opcode_ptr->op2, &is_var), 0, NULL);
 			xdebug_str_addc(&name, '[');
@@ -503,7 +482,7 @@ static char *xdebug_find_var_name(zend_execute_data *execute_data, const zend_op
 #if PHP_VERSION_ID >= 70400
 	if (cur_opcode->opcode == ZEND_ASSIGN_STATIC_PROP) {
 		dimval = xdebug_get_zval(execute_data, cur_opcode->op1_type, &cur_opcode->op1, &is_var);
-		xdebug_str_add(&name, xdebug_sprintf("%s", Z_STRVAL_P(dimval)), 1);
+		xdebug_str_add(&name, Z_STRVAL_P(dimval), 0);
 	}
 #endif
 
@@ -513,7 +492,7 @@ static char *xdebug_find_var_name(zend_execute_data *execute_data, const zend_op
 	return name.d;
 }
 
-static int xdebug_common_assign_dim_handler(const char *op, int do_cc, zend_execute_data *execute_data)
+static int xdebug_common_assign_dim_handler(const char *op, XDEBUG_OPCODE_HANDLER_ARGS)
 {
 	char    *file;
 	zend_op_array *op_array = &execute_data->func->op_array;
@@ -531,27 +510,16 @@ static int xdebug_common_assign_dim_handler(const char *op, int do_cc, zend_exec
 
 	/* TODO TEST FOR ASSIGNMENTS IN FILTERING */
 //	if (xdebug_is_top_stack_frame_filtered(XDEBUG_FILTER_CODE_COVERAGE)) {
-//		return ZEND_USER_OPCODE_DISPATCH;
+//		return xdebug_call_original_opcode_handler_if_set(cur_opcode->opcode, XDEBUG_OPCODE_HANDLER_ARGS_PASSTHRU);
 //	}
 
-	xdebug_coverage_record_assign_if_active(execute_data, op_array, do_cc);
-
-	if (XG_TRACE(trace_context) && XINI_BASE(collect_assignments)) {
+	if (XG_TRACE(trace_context) && XINI_TRACE(collect_assignments)) {
 		char *full_varname;
 
 		if (cur_opcode->opcode == ZEND_QM_ASSIGN && cur_opcode->result_type != IS_CV) {
-			return ZEND_USER_OPCODE_DISPATCH;
+			return xdebug_call_original_opcode_handler_if_set(cur_opcode->opcode, XDEBUG_OPCODE_HANDLER_ARGS_PASSTHRU);
 		}
-#if PHP_VERSION_ID < 70100
-		if (cur_opcode->opcode == ZEND_ASSIGN_REF) {
-			const zend_op *previous_opline = xdebug_find_referenced_opline(execute_data, cur_opcode, 1);
-			full_varname = xdebug_find_var_name(execute_data, previous_opline, cur_opcode);
-		} else {
-			full_varname = xdebug_find_var_name(execute_data, execute_data->opline, NULL);
-		}
-#else
 		full_varname = xdebug_find_var_name(execute_data, execute_data->opline, NULL);
-#endif
 
 		if (cur_opcode->opcode >= ZEND_PRE_INC && cur_opcode->opcode <= ZEND_POST_DEC) {
 			char *tmp_varname;
@@ -609,12 +577,7 @@ static int xdebug_common_assign_dim_handler(const char *op, int do_cc, zend_exec
 				right_full_varname = xdebug_sprintf("$%s", zend_get_compiled_variable_name(op_array, cur_opcode->op2.var)->val);
 			} else {
 				const zend_op *referenced_opline = xdebug_find_referenced_opline(execute_data, cur_opcode, 2);
-#if PHP_VERSION_ID < 70100
-				const zend_op *previous_opline = xdebug_find_referenced_opline(execute_data, cur_opcode, 1);
-				right_full_varname = xdebug_find_var_name(execute_data, referenced_opline, previous_opline + 1);
-#else
 				right_full_varname = xdebug_find_var_name(execute_data, referenced_opline, NULL);
-#endif
 			}
 #if PHP_VERSION_ID >= 70400
 		} else if (cur_opcode->opcode == ZEND_ASSIGN_OBJ_REF) {
@@ -636,55 +599,57 @@ static int xdebug_common_assign_dim_handler(const char *op, int do_cc, zend_exec
 			val = xdebug_get_zval(execute_data, cur_opcode->op2_type, &cur_opcode->op2, &is_var);
 		}
 
-		fse = XDEBUG_LLIST_VALP(XDEBUG_LLIST_TAIL(XG_BASE(stack)));
-		if (XG_TRACE(trace_context) && XINI_BASE(collect_assignments) && XG_TRACE(trace_handler)->assignment) {
+		fse = XDEBUG_VECTOR_TAIL(XG_BASE(stack));
+		if (XG_TRACE(trace_context) && XINI_TRACE(collect_assignments) && XG_TRACE(trace_handler)->assignment) {
 			XG_TRACE(trace_handler)->assignment(XG_TRACE(trace_context), fse, full_varname, val, right_full_varname, op, file, lineno);
 		}
 		xdfree(full_varname);
+		xdfree(right_full_varname);
 	}
-	return ZEND_USER_OPCODE_DISPATCH;
+
+	return xdebug_call_original_opcode_handler_if_set(cur_opcode->opcode, XDEBUG_OPCODE_HANDLER_ARGS_PASSTHRU);
 }
 
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign,"=",1)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(qm_assign,"=",1)
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign, "=");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(qm_assign, "=");
 #if PHP_VERSION_ID >= 70400
-XDEBUG_OPCODE_OVERRIDE_ASSIGN_OP(assign_op,0)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN_OP(assign_dim_op,0)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN_OP(assign_obj_op,0)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN_OP(assign_static_prop_op,0)
+XDEBUG_OPCODE_OVERRIDE_ASSIGN_OP(assign_op);
+XDEBUG_OPCODE_OVERRIDE_ASSIGN_OP(assign_dim_op);
+XDEBUG_OPCODE_OVERRIDE_ASSIGN_OP(assign_obj_op);
+XDEBUG_OPCODE_OVERRIDE_ASSIGN_OP(assign_static_prop_op);
 #else
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_add,"+=",0)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_sub,"-=",0)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_mul,"*=",0)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_div,"/=",0)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_mod,"%=",0)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_sl,"<<=",0)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_sr,">>=",0)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_bw_or,"|=",0)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_bw_and,"&=",0)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_bw_xor,"^=",0)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_pow,"**=",0)
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_add, "+=");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_sub, "-=");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_mul, "*=");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_div, "/=");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_mod, "%=");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_sl, "<<=");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_sr, ">>=");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_bw_or, "|=");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_bw_and, "&=");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_bw_xor, "^=");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_pow, "**=");
 #endif
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(pre_inc,"",0)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(post_inc,"",0)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(pre_dec,"",0)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(post_dec,"",0)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(pre_inc_obj,"",0)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(post_inc_obj,"",0)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(pre_dec_obj,"",0)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(post_dec_obj,"",0)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_concat,".=",1)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_dim,"=",1)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_obj,"=",1)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_ref,"=&",1)
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(pre_inc, "");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(post_inc, "");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(pre_dec, "");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(post_dec, "");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(pre_inc_obj, "");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(post_inc_obj, "");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(pre_dec_obj, "");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(post_dec_obj, "");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_concat, ".=");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_dim, "=");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_obj, "=");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_ref, "=&");
 #if PHP_VERSION_ID >= 70400
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_obj_ref,"=&",1)
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_static_prop, "=", 0);
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_static_prop_ref, "=&", 0);
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(pre_inc_static_prop, "", 0);
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(pre_dec_static_prop, "", 0);
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(post_inc_static_prop, "", 0);
-XDEBUG_OPCODE_OVERRIDE_ASSIGN(post_dec_static_prop, "", 0);
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_obj_ref, "=&");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_static_prop, "=");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(assign_static_prop_ref, "=&");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(pre_inc_static_prop, "");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(pre_dec_static_prop, "");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(post_inc_static_prop, "");
+XDEBUG_OPCODE_OVERRIDE_ASSIGN(post_dec_static_prop, "");
 #endif
 
 void xdebug_init_tracing_globals(xdebug_tracing_globals_t *xg)
@@ -696,47 +661,47 @@ void xdebug_init_tracing_globals(xdebug_tracing_globals_t *xg)
 void xdebug_tracing_minit(INIT_FUNC_ARGS)
 {
 	/* Override opcodes for variable assignments in traces */
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(include_or_eval, ZEND_INCLUDE_OR_EVAL);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(assign, ZEND_ASSIGN);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(qm_assign, ZEND_QM_ASSIGN);
+	xdebug_register_with_opcode_multi_handler(ZEND_ASSIGN, xdebug_assign_handler);
+	xdebug_register_with_opcode_multi_handler(ZEND_QM_ASSIGN, xdebug_qm_assign_handler);
+
 #if PHP_VERSION_ID >= 70400
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(assign_op, ZEND_ASSIGN_OP);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(assign_dim_op, ZEND_ASSIGN_DIM_OP);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(assign_obj_op, ZEND_ASSIGN_OBJ_OP);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(assign_static_prop_op, ZEND_ASSIGN_STATIC_PROP_OP);
+	xdebug_set_opcode_handler(ZEND_ASSIGN_OP, xdebug_assign_op_handler);
+	xdebug_set_opcode_handler(ZEND_ASSIGN_DIM_OP, xdebug_assign_dim_op_handler);
+	xdebug_set_opcode_handler(ZEND_ASSIGN_OBJ_OP, xdebug_assign_obj_op_handler);
+	xdebug_set_opcode_handler(ZEND_ASSIGN_STATIC_PROP_OP, xdebug_assign_static_prop_op_handler);
 #else
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(assign_add, ZEND_ASSIGN_ADD);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(assign_sub, ZEND_ASSIGN_SUB);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(assign_mul, ZEND_ASSIGN_MUL);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(assign_div, ZEND_ASSIGN_DIV);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(assign_mod, ZEND_ASSIGN_MOD);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(assign_pow, ZEND_ASSIGN_POW);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(assign_sl, ZEND_ASSIGN_SL);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(assign_sr, ZEND_ASSIGN_SR);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(assign_concat, ZEND_ASSIGN_CONCAT);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(assign_bw_or, ZEND_ASSIGN_BW_OR);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(assign_bw_and, ZEND_ASSIGN_BW_AND);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(assign_bw_xor, ZEND_ASSIGN_BW_XOR);
+	xdebug_set_opcode_handler(ZEND_ASSIGN_ADD, xdebug_assign_add_handler);
+	xdebug_set_opcode_handler(ZEND_ASSIGN_SUB, xdebug_assign_sub_handler);
+	xdebug_set_opcode_handler(ZEND_ASSIGN_MUL, xdebug_assign_mul_handler);
+	xdebug_set_opcode_handler(ZEND_ASSIGN_DIV, xdebug_assign_div_handler);
+	xdebug_set_opcode_handler(ZEND_ASSIGN_MOD, xdebug_assign_mod_handler);
+	xdebug_set_opcode_handler(ZEND_ASSIGN_POW, xdebug_assign_pow_handler);
+	xdebug_set_opcode_handler(ZEND_ASSIGN_SL, xdebug_assign_sl_handler);
+	xdebug_set_opcode_handler(ZEND_ASSIGN_SR, xdebug_assign_sr_handler);
+	xdebug_set_opcode_handler(ZEND_ASSIGN_CONCAT, xdebug_assign_concat_handler);
+	xdebug_set_opcode_handler(ZEND_ASSIGN_BW_OR, xdebug_assign_bw_or_handler);
+	xdebug_set_opcode_handler(ZEND_ASSIGN_BW_AND, xdebug_assign_bw_and_handler);
+	xdebug_set_opcode_handler(ZEND_ASSIGN_BW_XOR, xdebug_assign_bw_xor_handler);
 #endif
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(assign_dim, ZEND_ASSIGN_DIM);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(assign_obj, ZEND_ASSIGN_OBJ);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(assign_ref, ZEND_ASSIGN_REF);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(pre_inc, ZEND_PRE_INC);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(post_inc, ZEND_POST_INC);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(pre_dec, ZEND_PRE_DEC);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(post_dec, ZEND_POST_DEC);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(pre_inc_obj, ZEND_PRE_INC_OBJ);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(post_inc_obj, ZEND_POST_INC_OBJ);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(pre_dec_obj, ZEND_PRE_DEC_OBJ);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(post_dec_obj, ZEND_POST_DEC_OBJ);
+	xdebug_set_opcode_handler(ZEND_ASSIGN_DIM, xdebug_assign_dim_handler);
+	xdebug_set_opcode_handler(ZEND_ASSIGN_OBJ, xdebug_assign_obj_handler);
+	xdebug_set_opcode_handler(ZEND_ASSIGN_REF, xdebug_assign_ref_handler);
+	xdebug_set_opcode_handler(ZEND_PRE_INC, xdebug_pre_inc_handler);
+	xdebug_set_opcode_handler(ZEND_POST_INC, xdebug_post_inc_handler);
+	xdebug_set_opcode_handler(ZEND_PRE_DEC, xdebug_pre_dec_handler);
+	xdebug_set_opcode_handler(ZEND_POST_DEC, xdebug_post_dec_handler);
+	xdebug_set_opcode_handler(ZEND_PRE_INC_OBJ, xdebug_pre_inc_obj_handler);
+	xdebug_set_opcode_handler(ZEND_POST_INC_OBJ, xdebug_post_inc_obj_handler);
+	xdebug_set_opcode_handler(ZEND_PRE_DEC_OBJ, xdebug_pre_dec_obj_handler);
+	xdebug_set_opcode_handler(ZEND_POST_DEC_OBJ, xdebug_post_dec_obj_handler);
 #if PHP_VERSION_ID >= 70400
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(assign_obj_ref, ZEND_ASSIGN_OBJ_REF);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(assign_static_prop, ZEND_ASSIGN_STATIC_PROP);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(assign_static_prop_ref, ZEND_ASSIGN_STATIC_PROP_REF);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(pre_inc_static_prop, ZEND_PRE_INC_STATIC_PROP);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(pre_dec_static_prop, ZEND_PRE_DEC_STATIC_PROP);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(post_inc_static_prop, ZEND_POST_INC_STATIC_PROP);
-	XDEBUG_SET_OPCODE_OVERRIDE_ASSIGN(post_dec_static_prop, ZEND_POST_DEC_STATIC_PROP);
+	xdebug_set_opcode_handler(ZEND_ASSIGN_OBJ_REF, xdebug_assign_obj_ref_handler);
+	xdebug_set_opcode_handler(ZEND_ASSIGN_STATIC_PROP, xdebug_assign_static_prop_handler);
+	xdebug_set_opcode_handler(ZEND_ASSIGN_STATIC_PROP_REF, xdebug_assign_static_prop_ref_handler);
+	xdebug_set_opcode_handler(ZEND_PRE_INC_STATIC_PROP, xdebug_pre_inc_static_prop_handler);
+	xdebug_set_opcode_handler(ZEND_PRE_DEC_STATIC_PROP, xdebug_pre_dec_static_prop_handler);
+	xdebug_set_opcode_handler(ZEND_POST_INC_STATIC_PROP, xdebug_post_inc_static_prop_handler);
+	xdebug_set_opcode_handler(ZEND_POST_DEC_STATIC_PROP, xdebug_post_dec_static_prop_handler);
 #endif
 
 	REGISTER_LONG_CONSTANT("XDEBUG_TRACE_APPEND", XDEBUG_TRACE_OPTION_APPEND, CONST_CS | CONST_PERSISTENT);
@@ -753,22 +718,19 @@ void xdebug_tracing_rinit(void)
 
 void xdebug_tracing_post_deactivate(void)
 {
-	XG_TRACE(trace_context) = NULL;
-
 	if (XG_TRACE(trace_context)) {
 		xdebug_stop_trace();
 	}
+
+	XG_TRACE(trace_context) = NULL;
 }
 
 void xdebug_tracing_init_if_requested(zend_op_array *op_array)
 {
-	if (
-		(XINI_TRACE(auto_trace) || xdebug_trigger_enabled(XINI_TRACE(trace_enable_trigger), "XDEBUG_TRACE", XINI_TRACE(trace_enable_trigger_value)))
-		&& XINI_TRACE(trace_output_dir) && strlen(XINI_TRACE(trace_output_dir))
-	) {
+	if (xdebug_lib_start_with_request() || xdebug_lib_start_with_trigger(NULL)) {
 		/* In case we do an auto-trace we are not interested in the return
 		 * value, but we still have to free it. */
-		xdfree(xdebug_start_trace(NULL, STR_NAME_VAL(op_array->filename), XINI_TRACE(trace_options)));
+		xdfree(xdebug_start_trace(NULL, op_array->filename, XINI_TRACE(trace_options)));
 	}
 }
 
@@ -785,6 +747,8 @@ void xdebug_tracing_execute_ex(int function_nr, function_stack_entry *fse)
 
 void xdebug_tracing_execute_ex_end(int function_nr, function_stack_entry *fse, zend_execute_data *execute_data)
 {
+	zend_op_array *op_array;
+
 	if (fse->filtered_tracing || !XG_TRACE(trace_context)) {
 		return;
 	}
@@ -794,19 +758,23 @@ void xdebug_tracing_execute_ex_end(int function_nr, function_stack_entry *fse, z
 	}
 
 	/* Store return value in the trace file */
-	if (XINI_BASE(collect_return)) {
-		zend_op_array *op_array = &(execute_data->func->op_array);
+	if (!XINI_TRACE(collect_return)) {
+		return;
+	}
 
-		if (execute_data && execute_data->return_value) {
-			if (op_array->fn_flags & ZEND_ACC_GENERATOR) {
-				if (XG_TRACE(trace_handler)->generator_return_value) {
-					XG_TRACE(trace_handler)->generator_return_value(XG_TRACE(trace_context), fse, function_nr, (zend_generator*) execute_data->return_value);
-				}
-			} else {
-				if (XG_TRACE(trace_handler)->return_value) {
-					XG_TRACE(trace_handler)->return_value(XG_TRACE(trace_context), fse, function_nr, execute_data->return_value);
-				}
-			}
+	op_array = &(execute_data->func->op_array);
+
+	if (!execute_data || !execute_data->return_value) {
+		return;
+	}
+
+	if (op_array->fn_flags & ZEND_ACC_GENERATOR) {
+		if (XG_TRACE(trace_handler)->generator_return_value) {
+			XG_TRACE(trace_handler)->generator_return_value(XG_TRACE(trace_context), fse, function_nr, (zend_generator*) execute_data->return_value);
+		}
+	} else {
+		if (XG_TRACE(trace_handler)->return_value) {
+			XG_TRACE(trace_handler)->return_value(XG_TRACE(trace_context), fse, function_nr, execute_data->return_value);
 		}
 	}
 }
@@ -836,7 +804,7 @@ void xdebug_tracing_execute_internal_end(int function_nr, function_stack_entry *
 	}
 
 	/* Store return value in the trace file */
-	if (XINI_BASE(collect_return) && fse->function.type != XFUNC_ZEND_PASS && return_value && XG_TRACE(trace_handler)->return_value) {
+	if (XINI_TRACE(collect_return) && fse->function.type != XFUNC_ZEND_PASS && return_value && XG_TRACE(trace_handler)->return_value) {
 		XG_TRACE(trace_handler)->return_value(XG_TRACE(trace_context), fse, function_nr, return_value);
 	}
 }
